from flask import Flask
from flask_socketio import SocketIO
from sqlalchemy.orm import scoped_session
from hydrus.sync_namespace import SyncNamespace


socketio = SocketIO()


def create_socket(app: Flask, session: scoped_session) -> SocketIO:
    socketio.init_app(app, logger=True)
    socketio.on_namespace(SyncNamespace(namespace='/sync', db_session=session))
    return socketio



"""
Pluggable utilities for Hydrus.
===============================
Imports :
contextlib.contextmanager : This function is a decorator that can be used to
define a factory function for with statement context managers, without needing
to create a class or separate __enter__() and __exit__() methods.
Ref- https://docs.python.org/2/library/contextlib.html#contextlib.contextmanager
flask.appcontext_pushed : Signal is sent when an application context is pushed.
The sender is the application.
Ref- http://flask.pocoo.org/docs/0.12/api/#flask.appcontext_pushed
Ref- https://speakerdeck.com/mitsuhiko/advanced-flask-patterns-1
flask.g : Used to attach values to global variables
doc_writer_sample : Sample script used to create Hydra APIDocumentation
hydra_python_core.engine : An SQLalchemy DB engine
sqlalchemy.orm.sessionmaker : Used to create a SQLalchemy Session
sqlalchemy.orm.session.Session : SQLalchemy Session class
Ref- http://docs.sqlalchemy.org/en/latest/orm/session_basics.html
hydra_python_core.doc_writer.HydraDoc : Class for Hydra Documentation
"""  # nopep8

from contextlib import contextmanager
from flask import appcontext_pushed
from flask import g
from hydrus.samples import doc_writer_sample
from hydrus.data.db_models import engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.session import Session
from hydra_python_core.doc_writer import HydraDoc
from flask.app import Flask
from typing import Any, Iterator


@contextmanager
def set_authentication(application: Flask, authentication: bool) -> Iterator:
    """
    Set the whether API needs to be authenticated or not (before it is run in main.py).

    :param application: Flask app object
            <flask.app.Flask>
    :param authentication : Bool. API Auth needed or not
            <bool>


    Raises:
        TypeError: If `authentication` is not a boolean value.

    """
    if not isinstance(authentication, bool):
        raise TypeError("Authentication flag must be of type <bool>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.authentication_ = authentication
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_authentication() -> bool:
    """
    Check whether API needs to be authenticated or not.
    Return and sets False if not found.
    :return authentication : Bool. API Auth needed or not
            <bool>
    """
    try:
        authentication = getattr(g, 'authentication_')
    except AttributeError:
        authentication = False
        g.authentication_ = authentication
    return authentication


@contextmanager
def set_api_name(application: Flask, api_name: str) -> Iterator:
    """
    Set the server name or EntryPoint for the app (before it is run in main.py).
    :param application: Flask app object
            <flask.app.Flask>
    :param api_name : API/Server name or EntryPoint
            <str>

    Raises:
        TypeError: If `api_name` is not a string.

    """
    if not isinstance(api_name, str):
        raise TypeError("The api_name is not of type <str>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.api_name = api_name
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_api_name() -> str:
    """
    Get the server API name.
    Returns an sets "api" as api_name if not found.
    :return api_name : API/Server name or EntryPoint
            <str>
    """
    try:
        api_name = getattr(g, 'api_name')
    except AttributeError:
        api_name = "api"
        g.api_name = api_name
    return api_name


@contextmanager
def set_page_size(application: Flask, page_size: int) -> Iterator:
    """
    Set the page_size of a page view.
    :param application: Flask app object
            <flask.app.Flask>
    :param page_size : Number of maximum elements a page can contain
            <int>

    Raises:
        TypeError: If `page_size` is not an int.

    """
    if not isinstance(page_size, int):
        raise TypeError("The page_size is not of type <int>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.page_size = page_size
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_page_size() -> int:
    """
    Get the page_size of a page-view.
    :return page_size : Number of maximum elements a page view can contain.
            <int>
    """
    try:
        page_size = getattr(g, 'page_size')
    except AttributeError:
        page_size = 10
        g.page_size = page_size
    return page_size


@contextmanager
def set_pagination(application: Flask, paginate: bool) -> Iterator:
    """
    Enable or disable pagination.
    :param application: Flask app object
            <flask.app.Flask>
    :param paginate : Pagination enabled or not
            <bool>

    Raises:
        TypeError: If `paginate` is not a bool.

    """
    if not isinstance(paginate, bool):
        raise TypeError("The CLI argument 'pagination' is not of type <bool>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.paginate = paginate
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_pagination() -> bool:
    """
    Get the pagination status(Enable/Disable).
    :return paginate : Pagination enabled or not
            <bool>
    """
    try:
        paginate = getattr(g, 'paginate')
    except AttributeError:
        paginate = True
        g.paginate = paginate
    return paginate


@contextmanager
def set_doc(application: Flask, APIDOC: HydraDoc) -> Iterator:
    """
    Set the API Documentation for the app (before it is run in main.py).
    :param application: Flask app object
            <flask.app.Flask>
    :param APIDOC : Hydra Documentation object
            <hydra_python_core.doc_writer.HydraDoc>

    Raises:
        TypeError: If `APIDOC` is not an instance of `HydraDoc`.

    """
    if not isinstance(APIDOC, HydraDoc):
        raise TypeError(
            "The API Doc is not of type <hydra_python_core.doc_writer.HydraDoc>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.doc = APIDOC
    with appcontext_pushed.connected_to(handler, application):
        yield


@contextmanager
def set_token(application: Flask, token: bool) -> Iterator:
    """Set whether API needs to implement token based authentication.

    Raises:
        TypeError: If `token` is not a boolean value.

    """
    if not isinstance(token, bool):
        raise TypeError("Token flag must be of type <bool>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.token_ = token
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_doc() -> HydraDoc:
    """
    Get the server API Documentation.
    Returns and sets doc_writer_sample.api_doc if not found.
    :return apidoc : Hydra Documentation object
            <hydra_python_core.doc_writer.HydraDoc>
    """
    try:
        apidoc = getattr(g, 'doc')
    except AttributeError:
        g.doc = apidoc = doc_writer_sample.api_doc
    return apidoc


def get_token() -> bool:
    """Check wether API needs to be authenticated or not."""
    try:
        token = getattr(g, 'token_')
    except AttributeError:
        token = False
        g.token_ = token
    return token


@contextmanager
def set_hydrus_server_url(application: Flask, server_url: str) -> Iterator:
    """
    Set the server URL for the app (before it is run in main.py).
    :param application: Flask app object
            <flask.app.Flask>
    :param server_url : Server URL
            <str>

    Raises:
        TypeError: If the `server_url` is not a string.

    """
    if not isinstance(server_url, str):
        raise TypeError("The server_url is not of type <str>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.hydrus_server_url = server_url
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_hydrus_server_url() -> str:
    """
    Get the server URL.
    Returns and sets "http://localhost/" if not found.
    :return hydrus_server_url : Server URL
            <str>
    """
    try:
        hydrus_server_url = getattr(g, 'hydrus_server_url')
    except AttributeError:
        hydrus_server_url = "http://localhost/"
        g.hydrus_server_url = hydrus_server_url
    return hydrus_server_url


@contextmanager
def set_session(application: Flask, DB_SESSION: scoped_session) -> Iterator:
    """
    Set the Database Session for the app before it is run in main.py.
    :param application: Flask app object
            <flask.app.Flask>
    :param DB_SESSION: SQLalchemy Session object
            <sqlalchemy.orm.session.Session>

    Raises:
        TypeError: If `DB_SESSION` is not an instance of `scoped_session` or `Session`.

    """
    if not isinstance(DB_SESSION, scoped_session) and not isinstance(
            DB_SESSION, Session):
        raise TypeError(
            "The API Doc is not of type <sqlalchemy.orm.session.Session> or"
            " <sqlalchemy.orm.scoping.scoped_session>")

    def handler(sender: Flask, **kwargs: Any) -> None:
        g.dbsession = DB_SESSION
    with appcontext_pushed.connected_to(handler, application):
        yield


def get_session() -> scoped_session:
    """
    Get the Database Session from g.
    Returns and sets a default Session if not found
    :return session : SQLalchemy Session object
            <sqlalchemy.orm.scoped_session>
    """
    try:
        session = getattr(g, 'dbsession')
    except AttributeError:
        session = scoped_session(sessionmaker(bind=engine))
        g.dbsession = session
    return session


def get_collections_and_parsed_classes():
    """
    Get all the collections and parsed classes from
    the API doc.
    :return collections, parsed_classes
            <tuple>
    """
    api_doc = get_doc()
    parsed_classes = api_doc.parsed_classes
    collections = api_doc.collections
    return (collections, parsed_classes)



"""
This file (test_auth.py) contains the functional tests for the collection endpoints
with respect to authentication.
"""


import json

import pytest


def test_wrong_id_get(test_client, constants, headers_with_wrong_id, collection_names):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a GET request with wrong user id in header
    THEN check that a '401' or '400' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for collection_ in endpoints["collections"]:
        collection_name = collection_['@id'].split(f"{API_NAME}/")[1]
        if collection_name in collection_names:
            response = test_client.get(collection_['@id'])
            headers_with_wrong_id['X-Authentication'] = response.headers['X-Authentication']
            response_get = test_client.get(
                collection_['@id'], headers=headers_with_wrong_id)
            assert response_get.status_code == 401 or response_get.status_code == 400


def test_wrong_id_put(test_client, constants, put_allowed_class_names, headers_with_wrong_id):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a PUT request with wrong user id in header
    THEN check that a '401' or '400' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for endpoint in endpoints:
        if endpoint in put_allowed_class_names:
            response = test_client.get(endpoints[endpoint])
            headers_with_wrong_id['X-Authentication'] = response.headers['X-Authentication']
            response_put = test_client.put(endpoints[endpoint],
                                           headers=headers_with_wrong_id,
                                           data=json.dumps(dict(foo="bar")))
            assert response_put.status_code == 401 or response_put.status_code == 400


def test_wrong_pass_get(test_client, constants, collection_names, headers_with_wrong_pass):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a GET request with wrong user passphrase in header
    THEN check that a '401' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for collection_ in endpoints["collections"]:
        collection_name = collection_['@id'].split(f"{API_NAME}/")[1]
        if collection_name in collection_names:
            response = test_client.get(collection_['@id'])
            headers_with_wrong_pass['X-Authentication'] = response.headers['X-Authentication']
            response_get = test_client.get(
                collection_['@id'], headers=headers_with_wrong_pass)
            assert response_get.status_code == 401


def test_wrong_pass_put(test_client, constants, put_allowed_class_names, headers_with_wrong_pass):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a PUT request with wrong user passphrase in header
    THEN check that a '401' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for endpoint in endpoints:
        if endpoint in put_allowed_class_names:
            response = test_client.get(endpoints[endpoint])
            headers_with_wrong_pass['X-Authentication'] = response.headers['X-Authentication']
            response_put = test_client.put(endpoints[endpoint],
                                           headers=headers_with_wrong_pass,
                                           data=json.dumps(dict(foo="bar")))
            assert response_put.status_code == 401


def test_wrong_nonce_get(test_client, constants, collection_names,
                         headers_with_correct_pass_and_id):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a GET request with wrong nonce in header
    THEN check that a '401' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for collection_ in endpoints["collections"]:
        collection_name = collection_['@id'].split(f"{API_NAME}/")[1]
        if collection_name in collection_names:
            headers_with_correct_pass_and_id['X-authentication'] = "random-string"
            response_get = test_client.get(
                collection_['@id'], headers=headers_with_correct_pass_and_id)
            assert response_get.status_code == 401


def test_wrong_nonce_put(test_client, constants, put_allowed_class_names,
                         headers_with_correct_pass_and_id):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a PUT request with wrong nonce in header
    THEN check that a '401' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for endpoint in endpoints:
        if endpoint in put_allowed_class_names:
            headers_with_correct_pass_and_id['X-authentication'] = "random-string"
            response_put = test_client.put(endpoints[endpoint],
                                           headers=headers_with_correct_pass_and_id,
                                           data=json.dumps(dict(foo="bar")))
            assert response_put.status_code == 401


# get and delete can be parametrize together as for both of them no data object in request
@pytest.mark.parametrize("operation", [
    ("get"),
    ("delete"),
])
def test_correct_auth_get(operation, test_client, constants, collection_names,
                          headers_with_correct_pass_and_id):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a GET or a DELETE request with correct user credentials
    THEN check that a '200' or '405' status code is returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for collection_ in endpoints["collections"]:
        collection_name = collection_['@id'].split(f"{API_NAME}/")[1]
        if collection_name in collection_names:
            response = test_client.get(collection_['@id'])
            x_auth = 'X-Authentication'
            headers_with_correct_pass_and_id[x_auth] = response.headers[x_auth]
            # get the response for the required operation
            response_op = getattr(test_client, operation)(collection_['@id'],
                                                          headers=headers_with_correct_pass_and_id)
            assert response_op.status_code == 200 or response_op.status_code == 405


# post and put can be parametrized together as for them data object in request can be same
@pytest.mark.parametrize("operation", [
    ("put"),
])
def test_correct_auth_post(operation, test_client, constants, put_allowed_class_names,
                           headers_with_correct_pass_and_id):
    """
    GIVEN a Flask application
    WHEN a collection endpoint has a PUT request with correct user credentials
    THEN check that a '401' status code is not returned
    """
    API_NAME = constants["API_NAME"]
    response = test_client.get(f"/{API_NAME}")
    endpoints = json.loads(response.data.decode('utf-8'))
    for endpoint in endpoints:
        if endpoint in put_allowed_class_names:
            response = test_client.get(endpoints[endpoint])
            x_auth = 'X-Authentication'
            headers_with_correct_pass_and_id[x_auth] = response.headers[x_auth]
            response_op = getattr(test_client, operation)(endpoints[endpoint],
                                                          headers=headers_with_correct_pass_and_id,
                                                          data=json.dumps(dict(foo="bar")))
            assert response_op.status_code != 401



from threading import Timer
from hydrus.data.db_models import Modification
from sqlalchemy.orm import scoped_session


def remove_stale_modification_records(session: scoped_session,
                                      stale_records_removal_interval: int=900):
    """
    Remove modification records which are older than last 1000 records.
    :param session: sqlalchemy session.
    :param stale_records_removal_interval: Interval time to run the removal job.
    """
    timer = Timer(stale_records_removal_interval,
                  remove_stale_modification_records, [session])
    timer.daemon = True
    timer.start()
    # Get all valid records.
    valid_records = session.query(Modification).order_by(
        Modification.job_id.desc()).limit(1000).all()
    # If number of returned valid records is less than set limit then
    # there is nothing to clean up.
    if len(valid_records) < 1000:
        return
    else:
        # Get the job_id of last (oldest) valid record.
        job_id_of_last_valid_record = valid_records[-1].job_id
        # Get all records which are older than the oldest valid record.
        stale_records = session.query(Modification).filter(
            Modification.job_id < job_id_of_last_valid_record).all()
        for record in stale_records:
            session.delete(record)
        session.commit()
    session.remove()



# from sqlalchemy.orm.session import Session

from typing import Dict, Any, Tuple
from hydrus.data.exceptions import (
    PageNotFound,
    IncompatibleParameters,
    OffsetOutOfRange,
)
from hydrus.conf import get_host_domain


def recreate_iri(API_NAME: str, path: str, search_params: Dict[str, Any]) -> str:
    """Recreate the IRI with query arguments
    :param API_NAME: API name specified while starting the server.
    :param path: endpoint
    :param search_params: List of query parameters.
    :return: Recreated IRI.
    """
    iri = f"/{API_NAME}/{path}?"
    for param in search_params:
        # Skip page, pageIndex or offset parameters as they will be updated to point to
        # next, previous and last page
        if param in ["page", "pageIndex", "offset"]:
            continue
        iri += f"{param}={search_params[param]}&"
    return iri


def parse_search_params(search_params: Dict[str, Any]) -> Dict[str, Any]:
    """Parse nested search parameters and add them in the form of nested dict.
    :param search_params: Dictionary having input search parameters.
    :return: A dictionary having keys as nested dictionaries for each nested search param
    """
    for param in search_params.copy():
        # For one level deep nested parameters
        if "[" in param and "]" in param:
            value = search_params[param]
            split_param_list = param.split('[')
            prop_name = split_param_list[0]
            nested_prop_name = split_param_list[1].split(']')[0]
            if prop_name not in search_params:
                search_params[prop_name] = {}
            search_params[prop_name][nested_prop_name] = value
            search_params.pop(param)
        # For normal parameters
        else:
            # Skip if the parameter is not nested
            continue
    return search_params


def calculate_page_limit_and_offset(paginate: bool, page_size: int, page: int, result_length: int,
                                    offset: int, limit: int) -> Tuple[int, int]:
    """Calculate page limit and offset for pagination.
    :param paginate: Showing whether pagination is enable/disable.
    :param page_size: Number maximum elements showed in a page.
    :param page: page number.
    :param result_length: Length of the list containing desired elements.
    :param offset: offset value.
    :param limit: page limit.
    :return: page limit and offset.
    """
    if limit is not None:
        page_size = limit
    if paginate is True:
        if offset is None:
            offset = (page - 1) * page_size
        limit = page_size
    else:
        offset = 0
        limit = result_length

    return limit, offset


def pre_process_pagination_parameters(search_params: Dict[str, Any], paginate: bool,
                                      page_size: int, result_length: int) -> Tuple[int, int, int]:
    """Pre-process pagination related query parameters passed by client.
    :param search_params: Dict of all search parameters.
    :param paginate: Indicates if pagination is enabled/disabled.
    :param page_size: Maximum element a page can contain.
    :param result_length: Length of the list of containing desired items.
    :return: returns page number, page limit and offset.
    """
    incompatible_parameters = ["page", "pageIndex", "offset"]
    incompatible_parameters_len = len(incompatible_parameters)
    # Find any pair of incompatible parameters
    for i in range(incompatible_parameters_len):
        if incompatible_parameters[i] not in search_params:
            continue
        if i != incompatible_parameters_len - 1:
            for j in range(i+1, incompatible_parameters_len):
                if incompatible_parameters[j] in search_params:
                    raise IncompatibleParameters([incompatible_parameters[i],
                                                  incompatible_parameters[j]])
    try:
        # Extract page number from query arguments
        if "pageIndex" in search_params:
            page = int(search_params.get("pageIndex"))
            offset = None
        elif "offset" in search_params:
            offset = int(search_params.get("offset"))
            page = offset//page_size + 1
            if offset > result_length:
                raise OffsetOutOfRange(str(offset))
        else:
            page = int(search_params.get("page", 1))
            offset = None
        if "limit" in search_params:
            limit = int(search_params.get("limit"))
        else:
            limit = None
    except ValueError:
        raise PageNotFound(page)
    page_limit, offset = calculate_page_limit_and_offset(paginate=paginate, page_size=page_size,
                                                         page=page, result_length=result_length,
                                                         offset=offset, limit=limit)
    return page, page_limit, offset


def attach_hydra_view(collection_template: Dict[str, Any], paginate_param: str,
                      result_length: int, page_size: int,
                      iri: str, offset: int = None,
                      page: int = None, last: int = None) -> None:
    """Attaches hydra:view to the collection template.
    :param collection_template: the collection template.
    :param paginate_param: type of paginate parameter used.
    :param result_length: length of the result set.
    :param page_size: size of the page.
    :param iri: IRI of the collection with query parameters except "page", "pageIndex" and "offset".
    :param offset: offset used for pagination, None if not used.
    :param page: page number used for pagination, None if not used.
    :param last: Page number of the last page only used when "page" or "pageIndex"
                 is used for pagination, None otherwise.
    """
    if paginate_param == "offset":
        collection_template["hydra:view"] = {
            "@id": f"{get_host_domain()}{iri}{paginate_param}={offset}",
            "@type": "hydra:PartialCollectionView",
            "hydra:first": f"{get_host_domain()}{iri}{paginate_param}=0",
            "hydra:last": f"{get_host_domain()}{iri}{paginate_param}={result_length - page_size}"
        }
        if offset > page_size:
            collection_template["hydra:view"]["hydra:previous"] = (
                f"{iri}{paginate_param}={offset - page_size}")
        if offset < result_length-page_size:
            collection_template["hydra:view"]["hydra:next"] = (
                f"{iri}{paginate_param}={offset + page_size}")
    else:
        collection_template["hydra:view"] = {
            "@id": f"{get_host_domain()}{iri}{paginate_param}={page}",
            "@type": "hydra:PartialCollectionView",
            "hydra:first": f"{get_host_domain()}{iri}{paginate_param}=1",
            "hydra:last": f"{get_host_domain()}{iri}{paginate_param}={last}"
        }
        if page != 1:
            collection_template["hydra:view"]["hydra:previous"] = (
                f"{iri}{paginate_param}={page-1}")
        if page != last:
            collection_template["hydra:view"]["hydra:next"] = (
               f"{iri}{paginate_param}={page + 1}")



import random
import string
import uuid
from base64 import b64encode

import pytest
from hydra_python_core import doc_maker
from hydra_python_core.doc_writer import DocUrl, HydraLink
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from hydrus.app_factory import app_factory
from hydrus.data import crud, doc_parse
from hydrus.data.db_models import Base, create_database_tables
from hydrus.data.user import add_user
from hydrus.helpers import get_path_from_type
from hydrus.samples import doc_writer_sample, hydra_doc_sample
from hydrus.socketio_factory import create_socket
from hydrus.utils import (get_api_name, get_session, set_api_name,
                          set_authentication, set_doc, set_page_size,
                          set_session, set_token)


def get_doc_classes_and_properties(doc):
    """
    Extract classes and properties from a given HydraDoc object

    :param doc: HydraDoc object whose classes and properties have to extracted
    :type doc: HydraDoc
    :return: classes and properties in the HydraDoc object in a tuple
    :rtype: tuple(list, set)
    """
    test_classes = doc_parse.get_classes(doc)
    test_properties = doc_parse.get_all_properties(test_classes)
    return (test_classes, test_properties)


def gen_dummy_object(class_title, doc):
    """
    Create a dummy object based on the definitions in the API Doc.
    :param class_title: Title of the class whose object is being created.
    :param doc: ApiDoc.
    :return: A dummy object of class `class_title`.
    """
    object_ = {
        "@type": class_title
    }
    expanded_base_url = DocUrl.doc_url
    for class_path in doc.collections:
        if class_title == doc.collections[class_path]["collection"].name:
            members = list()
            manages_class_titles = list()
            collection_manages = doc.collections[class_title]["collection"].manages
            if type(collection_manages) is dict:
                # only one manages block
                manages_class = collection_manages['object'].split(expanded_base_url)[1]
                manages_class_titles.append(manages_class)
            elif type(collection_manages) is list:
                # multiple manages block
                for manages_block in collection_manages:
                    manages_class = collection_manages['object'].split(expanded_base_url)[1]
                    manages_class_titles.append(manages_class)
            for _ in range(3):
                member_class = random.choice(manages_class_titles)
                member = gen_dummy_object(member_class, doc)
                member_id = crud.insert(object_=member,
                                        session=get_session(),
                                        collection=False)
                member_class_path = get_path_from_type(member_class)
                member_api_path = f'/{get_api_name()}/{member_class_path}/{member_id}'
                members.append({
                    "@id": member_api_path,
                    "@type": member_class,
                })
            object_['members'] = members
            return object_
    for class_path in doc.parsed_classes:
        if class_title == doc.parsed_classes[class_path]["class"].title:
            for prop in doc.parsed_classes[class_path]["class"].supportedProperty:
                if prop.write is False:
                    continue
                if isinstance(prop.prop, HydraLink):
                    object_[prop.title] = ''.join(random.choice(
                        string.ascii_uppercase + string.digits) for _ in range(6))
                    pass
                elif expanded_base_url in prop.prop:
                    prop_class = prop.prop.split(expanded_base_url)[1]
                    object_[prop.title] = gen_dummy_object(prop_class, doc)
                else:
                    object_[prop.title] = ''.join(random.choice(
                        string.ascii_uppercase + string.digits) for _ in range(6))
            return object_


@pytest.fixture(scope='module')
def constants():
    """
    Constant values to be used in all tests
    """
    return {
        'API_NAME': 'demoapi',
        'HYDRUS_SERVER_URL': 'http://hydrus.com/',
        'PAGE_SIZE': 1
    }


@pytest.fixture(scope='module', name='doc')
def test_doc(constants):
    """
    Generate a test HydraDoc object from a Api Documentation
    """
    HYDRUS_SERVER_URL = constants['HYDRUS_SERVER_URL']
    API_NAME = constants['API_NAME']

    doc = doc_maker.create_doc(doc_writer_sample.api_doc.generate(),
                               HYDRUS_SERVER_URL,
                               API_NAME)
    return doc


@pytest.fixture(scope='module')
def session(engine):
    """
    Initialize a flask scoped session binded with a database
    """
    session = scoped_session(sessionmaker(bind=engine))
    yield session
    session.close()


@pytest.fixture(scope='module')
def engine():
    """
    Initialize a sqlalchemy engine binded with a database
    """
    engine = create_engine('sqlite:///:memory:')
    return engine


@pytest.fixture(scope='module')
def init_db_for_auth_tests(doc, session, engine):
    """
    Initialize the database by adding the classes and properties of
    test HydraDoc object and adding a user to the database
    """
    test_classes, test_properties = get_doc_classes_and_properties(doc)
    Base.metadata.create_all(engine)
    add_user(1, 'test', session)


@pytest.fixture()
def headers_with_wrong_id():
    """
    Get headers for wrong id and correct user paraphrase in credentials
    """
    credentials_with_wrong_id = b64encode(b"2:test").decode("ascii")
    headers = {"X-Authentication": "",
               "Authorization": f"Basic {credentials_with_wrong_id}"}
    return headers


@pytest.fixture()
def headers_with_wrong_pass():
    """
    Get headers for wrong user paraphrase and correct user id in credentials
    """
    credentials_with_wrong_pass = b64encode(b"1:test2").decode("ascii")
    headers = {"X-Authentication": "",
               "Authorization": f"Basic {credentials_with_wrong_pass}"}
    return headers


@pytest.fixture()
def headers_with_correct_pass_and_id():
    """
    Get headers for correct paraphrase and id in credentials
    """
    correct_credentials = b64encode(b"1:test").decode("ascii")
    headers = {"X-Authentication": "",
               "Authorization": f"Basic {correct_credentials}"}
    return headers


@pytest.fixture(scope='module')
def collection_names(doc):
    """
    Get the names of all the collection endpoints in a given HydraDoc object
    """
    return [collection_object['collection'].name
            for collection_object in doc.collections.values()]


@pytest.fixture(scope='module')
def put_allowed_class_names(doc):
    """
    Get the names of all the parsed classes endpoints in a given HydraDoc object
    """

    allowed_classes = list()
    for parsed_class in doc.parsed_classes.values():
        for operation in parsed_class['class'].supportedOperation:
            if operation.method == 'PUT':
                allowed_classes.append(parsed_class['class'].title)
    return allowed_classes


@pytest.fixture(scope='module', name='test_client')
def test_client_for_auth_tests(constants, session, doc, init_db_for_auth_tests, app):
    """
    Get a test flask app for testing
    """
    API_NAME = constants['API_NAME']
    with set_authentication(app, True):
        with set_token(app, False):
            with set_api_name(app, API_NAME):
                with set_doc(app, doc):
                    with set_session(app, session):
                        testing_client = app.test_client()
                        # Establish an application context before running the tests.
                        ctx = app.app_context()
                        ctx.push()
                        yield testing_client
                        ctx.pop()


@pytest.fixture(scope='module')
def drone_doc(constants):
    API_NAME = constants['API_NAME']
    HYDRUS_SERVER_URL = constants['HYDRUS_SERVER_URL']
    doc = doc_maker.create_doc(hydra_doc_sample.doc, HYDRUS_SERVER_URL, API_NAME)
    return doc


@pytest.fixture(scope='module')
def drone_doc_parsed_classes(drone_doc):
    return [drone_doc.parsed_classes[i]['class'].title for i in drone_doc.parsed_classes]


@pytest.fixture(scope='module')
def init_db_for_crud_tests(drone_doc, session, engine):
    """
    Initialize the database by adding the classes and properties of
    Drone Api test HydraDoc object.
    """
    test_classes, test_properties = get_doc_classes_and_properties(drone_doc)
    create_database_tables(test_classes)
    Base.metadata.create_all(engine)


@pytest.fixture(scope='module')
def app(constants):
    """
    Get a test flask app for testing in test
    """
    API_NAME = constants['API_NAME']
    app = app_factory(API_NAME)
    return app


@pytest.fixture(scope='module', name='test_app_client')
def test_client_for_app_tests(app, session, constants, doc):
    API_NAME = constants['API_NAME']
    PAGE_SIZE = constants['PAGE_SIZE']
    with set_api_name(app, API_NAME):
        with set_session(app, session):
            with set_doc(app, doc):
                with set_page_size(app, PAGE_SIZE):
                    testing_client = app.test_client()
                    # Establish an application context before running the tests.
                    ctx = app.app_context()
                    ctx.push()
                    yield testing_client
                    ctx.pop()


@pytest.fixture()
def init_db_for_app_tests(doc, constants, session, add_doc_classes_and_properties_to_db):
    """
    Initalze the database for testing app in
    tests/functional/test_app.py.
    """
    for class_ in doc.parsed_classes:
        class_title = doc.parsed_classes[class_]['class'].title
        dummy_obj = gen_dummy_object(class_title, doc)
        crud.insert(dummy_obj, id_=str(uuid.uuid4()), session=session)
        # If it's a collection class then add an extra object so
        # we can test pagination thoroughly.
        if class_ in doc.collections:
            crud.insert(dummy_obj, id_=str(uuid.uuid4()), session=session)


@pytest.fixture()
def init_db_for_socket_tests(doc, add_doc_classes_and_properties_to_db, session):
    """
    Initalze the database for testing app in
    tests/functional/test_socket.py.
    """
    for class_ in doc.parsed_classes:
        class_title = doc.parsed_classes[class_]["class"].title
        dummy_obj = gen_dummy_object(class_title, doc)
        crud.insert(dummy_obj, id_=str(uuid.uuid4()), session=session)
        # If it's a collection class then add an extra object so
        # we can test pagination thoroughly.
        if class_ in doc.collections:
            crud.insert(dummy_obj, id_=str(uuid.uuid4()), session=session)
    # Add two dummy modification records
    crud.insert_modification_record(method="POST", resource_url="", session=session)
    crud.insert_modification_record(method="DELETE", resource_url="", session=session)


@pytest.fixture(scope='module')
def socketio(app, session):
    socket = create_socket(app, session)
    return socket


@pytest.fixture(scope='module')
def socketio_client(app, session, constants, doc, socketio):
    API_NAME = constants['API_NAME']
    PAGE_SIZE = constants['PAGE_SIZE']
    with set_api_name(app, API_NAME):
        with set_session(app, session):
            with set_doc(app, doc):
                with set_page_size(app, PAGE_SIZE):
                    socketio_client = socketio.test_client(app, namespace='/sync')
                    return socketio_client


@pytest.fixture(scope='module')
def add_doc_classes_and_properties_to_db(doc, session, engine):
    """
    Add the doc classes and properties to database
    for testing in /functional/test_app.py and
    /functional/test_socket.py
    """
    test_classes, test_properties = get_doc_classes_and_properties(doc)
    # temporarily add manages block explicitly to collection-classes
    # until appropriate changes take place in hydra-python-core library
    manages = {
        "object": "vocab:dummyClass",
        "property": "rdf:type"
    }
    for class_ in test_classes:
        if "Collection" in class_['@id']:
            class_['manages'] = manages

    try:
        create_database_tables(test_classes)
    except Exception:
        # catch error when the tables have been already defined.
        # happens when /test_socket.py is run after /test_app.py
        # in the same session
        # in that case, no need to create the tables again on the
        # same sqlalchemy.ext.declarative.declarative_base instance
        pass
    Base.metadata.create_all(engine)



"""Helper functions for Item Collection"""

import json
from functools import partial

from flask import Response, jsonify, request, abort
from hydra_python_core.doc_writer import HydraStatus, HydraError, DocUrl

from hydrus.data import crud
from hydrus.data.exceptions import (
    ClassNotFound,
    InstanceExists,
    PropertyNotFound,
    PageNotFound,
    InvalidSearchParameter,
    OffsetOutOfRange,
    PropertyNotGiven
)

from hydrus.helpers import (
    set_response_headers,
    getType,
    hydrafy,
    add_iri_template,
    error_response,
    validate_object,
    parse_collection_members,
)

from hydrus.utils import (
    get_session,
    get_api_name,
    get_hydrus_server_url,
    get_page_size,
    get_pagination,
    get_collections_and_parsed_classes
)


def item_collection_get_response(path: str) -> Response:
    """
    Handles GET operation on item collection classes.

    :param path: Path for Item Collection
    :type path: str
    :return: Appropriate response for the GET operation.
    :rtype: Response
    """
    search_params = request.args.to_dict()
    collections, parsed_classes = get_collections_and_parsed_classes()
    api_name = get_api_name()
    expanded_base_url = DocUrl.doc_url
    # If endpoint and GET method is supported in the API and class is supported
    if path in parsed_classes:
        abort(405)
    if path in collections:
        collection = collections[path]["collection"]
        class_name = collection.manages["object"].split(expanded_base_url)[1]
        collection_manages_class = parsed_classes[class_name]["class"]
        class_type = collection_manages_class.title
        class_path = collection_manages_class.path
    try:
        # Get collection details from the database
        # create partial function for crud operation
        crud_response = partial(crud.get_collection, api_name,
                                class_type, session=get_session(),
                                path=path, search_params=search_params,
                                collection=False)
        if get_pagination():
            # Get paginated response
            response = crud_response(paginate=True,
                                     page_size=get_page_size())
        else:
            # Get whole collection
            response = crud_response(paginate=False)

        response["search"] = add_iri_template(path=class_path,
                                              API_NAME=api_name,
                                              collection_path=path)

        return set_response_headers(jsonify(hydrafy(response, path=path)))

    except (ClassNotFound, PageNotFound, InvalidSearchParameter,
            OffsetOutOfRange) as e:
        error = e.get_HTTP()
        return error_response(error)


def item_collection_put_response(path: str) -> Response:
    """
    Handles PUT operation on item collection classes.

    :param path: Path for Item Collection
    :type path: str
    :return: Appropriate response for the PUT operation.
    :rtype: Response
    """
    object_ = json.loads(request.data.decode('utf-8'))
    collections, parsed_classes = get_collections_and_parsed_classes()
    is_collection = False
    if path in parsed_classes:
        class_path = path
        is_collection = False
        obj_type = getType(path, "PUT")
    elif path in collections:
        collection = collections[path]["collection"]
        class_path = collection.path
        obj_type = collection.name
        is_collection = True
    if validate_object(object_, obj_type, class_path):
        # If Item in request's JSON is a valid object ie. @type is a key in object_
        # and the right Item type is being added to the collection
        if is_collection:
            object_ = parse_collection_members(object_)
        try:
            # Insert object and return location in Header
            object_id = crud.insert(object_=object_, session=get_session(),
                                    collection=is_collection)
            resource_url = f"{get_hydrus_server_url()}{get_api_name()}/{path}/{object_id}"
            headers_ = [{"Location": resource_url}]
            status_description = f"Object with ID {object_id} successfully added"
            status = HydraStatus(code=201,
                                 title="Object successfully added",
                                 desc=status_description)
            status_response = status.generate()
            status_response["iri"] = resource_url
            return set_response_headers(
                jsonify(status_response), headers=headers_,
                status_code=status.code)
        except (ClassNotFound, InstanceExists, PropertyNotFound,
                PropertyNotGiven) as e:
            error = e.get_HTTP()
            return error_response(error)
    else:
        error = HydraError(code=400, title="Data is not valid")
        return error_response(error)



"""Basic CRUD operations for the server.

    ===============================
    Imports :

    sqlalchemy.orm.with_polymorphic : Load columns for inheriting classes.
    Ref : http://docs.sqlalchemy.org/en/latest/orm/query.html

    sqlalchemy.exists : A convenience method that turns a query into an EXISTS subquery
    of the form EXISTS (SELECT 1 FROM … WHERE …).
    Ref : http://docs.sqlalchemy.org/en/latest/orm/query.html

    sqlalchemy.orm.exc.NoResultFound : A database result was required but none was found.
    Ref : http://docs.sqlalchemy.org/en/latest/orm/exceptions.html?highlight=result%20found#sqlalchemy.orm.exc.NoResultFound

    sqlalchemy.orm.session.Session : Manages persistence operations for ORM-mapped objects.
    Ref : http://docs.sqlalchemy.org/en/latest/orm/session_api.html?highlight=session#module-sqlalchemy.orm.session

    hydrus.data.db_models.Graph : Model for a graph that store triples of instance from the other models to map relationships.
    hydrus.data.db_models.BaseProperty : Model for Basic Property.
    hydrus.data.db_models.RDFClass : Model for Classes specifically RDF-OWL or RDF-HYDRA classes.
    hydrus.data.db_models.Instance : Model for Object/Resource. Instances are instances of some kind/classes that are served through the API.
    hydrus.data.db_models.Terminal : Model for Terminals which are numbers or string that can be referenced by a Property.
    hydrus.data.db_models.GraphIAC : Graph model for Instance >> AbstractProperty >> Class.
    hydrus.data.db_models.GraphIIT : Graph model for Instance >> InstanceProperty >> Terminal.
    hydrus.data.db_models.GraphIII : Graph model for Instance >> InstanceProperty >> Instance.

    Ref : ./db_models.py

    hydrus.data.exceptions : Contains all exceptions .
    typing : Module which provides support for type hints .

"""  # nopep8
import copy
from sqlalchemy.orm.exc import NoResultFound
from hydrus.data.db_models import Modification
from hydrus.data.exceptions import (
    InstanceExists,
    PageNotFound,
    IncompatibleParameters,
    OffsetOutOfRange)
from hydrus.data.crud_helpers import (
    recreate_iri,
    attach_hydra_view,
    pre_process_pagination_parameters,
    parse_search_params
)
# from sqlalchemy.orm.session import Session
from sqlalchemy.orm.scoping import scoped_session
from typing import Dict, Optional, Any, List

from hydrus.data.resource_based_classes import (
    get_object,
    insert_object,
    update_object,
    delete_object,
    get_all_filtered_instances,
    get_single_response,
    get_database_class,
    get_collection_member,
    delete_collection_member
)
from hydrus.conf import get_host_domain


def get(id_: str, type_: str, api_name: str, session: scoped_session,
        path: str = None, collection: bool = False) -> Dict[str, str]:
    """Retrieve an Instance with given ID from the database [GET].
    :param id_: id of object to be fetched
    :param type_: type of object
    :param api_name: name of api specified while starting server
    :param session: sqlalchemy scoped session
    :param path: endpoint
    :param collection: True if the type_ is of a collection, False for any other class
    :return: response to the request


    Raises:
        ClassNotFound: If the `type_` is not a valid/defined RDFClass.
        InstanceNotFound: If no Instance of the 'type_` class if found.

    """
    query_info = {
        "@type": type_,
        "id_": id_
    }

    object_template = get_object(query_info, session, collection)
    object_template["@id"] = f"{get_host_domain()}/{api_name}/{path}/{id_}"

    return object_template


def insert(object_: Dict[str, Any], session: scoped_session, id_: Optional[str] = None,
           collection: bool = False) -> str:
    """Insert an object to database [POST] and returns the inserted object.
    :param object_: object to be inserted
    :param session: sqlalchemy scoped session
    :param id_: id of the object to be inserted (optional param)
    :param collection: True if the type_ is of a collection, False for any other class

    :return: ID of object inserted


    Raises:
        ClassNotFound: If `object_["@type"] is not a valid/defined RDFClass.
        InstanceExists: If an Instance `id_` already exists.
        PropertyNotFound: If any property name of `object_` other than `@type` or `@context`
            is not a valid/defined property.
        NotInstanceProperty: If any property of `object_` is a dictionary but
            not an Instance property
        NotAbstractProperty: If any property of `object_` is a
            valid/defined RDFClass but is not a dictionary neither an Abstract Property
    """
    object_template = copy.deepcopy(object_)
    if id_ is not None:
        object_template['id'] = id_
    inserted_object_id = insert_object(object_template, session, collection)
    return inserted_object_id


def insert_multiple(objects_: List[Dict[str,
                                        Any]],
                    session: scoped_session,
                    id_: Optional[str] = "") -> List[str]:
    """
    Adds a list of object with given ids to the database
    :param objects_: List of dict's to be added to the database
    :param session: scoped session from getSession in utils
    :param id_: optional parameter containing the ids of objects that have to be inserted
    :return: Ids that have been inserted

    Raises:
        ClassNotFound: If any dict of `objects_` is not a valid/defined RDFClass.
        InstanceExists: If an Instance with same id already exists.
        PropertyNotFound: If for any dict in 'objects_' if any property is not
            a valid/defined property.
        NotAnInstanceProperty: If any property of a dict in `object_` is a dictionary but
            not an Instance property
        NotAnAbstractProperty: If any property of a dict in `object_` is a
            valid/defined RDFClass but is not a dictionary neither an Abstract Property

    """
    # import pdb;pdb.set_trace()

    id_list = id_.split(',')

    # list to hold all the ids of inserted objects
    instance_id_list = []

    for index in range(len(objects_)):
        id_of_object_ = None
        object_ = objects_[index]
        # check if id_ exist for object at that index
        try:
            id_of_object_ = id_list[index]
        except IndexError:
            pass
        inserted_object_id = insert(object_, session, id_of_object_)
        instance_id_list.append(inserted_object_id)

    return instance_id_list


def delete(id_: str, type_: str, session: scoped_session, collection: bool = False) -> None:
    """Delete an Instance and all its relations from DB given id [DELETE].
    :param id_: id of object to be deleted
    :param type_: type of object to be deleted
    :param session: sqlalchemy scoped session

    Raises:
        ClassNotFound: If `type_` does not represent a valid/defined RDFClass.
        InstanceNotFound: If no instace of type `type_` with id `id_` exists.

    """
    query_info = {
        "@type": type_,
        "id_": id_
    }
    delete_object(query_info, session, collection)


def delete_multiple(
        id_: List[int],
        type_: str,
        session: scoped_session) -> None:
    """
    To delete multiple rows in a single request
    :param id_: list of ids for objects to be deleted\
    :param type_: type of object to be deleted
    :param session: sqlalchemy scoped session

    Raises:
        ClassNotFound: If `type_` does not represent a valid/defined RDFClass.
        InstanceNotFound: If any instance with type 'type_' and any id in 'id_' list
            does not exist.

    """
    id_list = id_.split(',')
    for object_id_ in id_list:
        delete(object_id_, type_, session)


def update(id_: str,
           type_: str,
           object_: Dict[str,
                         str],
           session: scoped_session,
           api_name: str,
           path: str = None,
           collection: bool = False) -> str:
    """Update an object properties based on the given object [PUT].
    :param id_: if of object to be updated
    :param type_: type of object to be updated
    :param object_: object that has to be inserted
    :param session: sqlalchemy scoped session
    :param api_name: api name specified while starting server
    :param path: endpoint
    :param collection: True if the type_ is of a collection, False for any other class
    :return: id of updated object
    """
    query_info = {
        "@type": type_,
        "id_": id_
    }
    updated_object_id = update_object(object_, query_info, session, collection)
    return updated_object_id


def get_collection(API_NAME: str,
                   type_: str,
                   session: scoped_session,
                   paginate: bool,
                   page_size: int,
                   search_params: Dict[str, Any]=None,
                   path: str = None,
                   collection: bool = False) -> Dict[str, Any]:
    """Retrieve a type of collection from the database.
    :param API_NAME: api name specified while starting server
    :param type_: type of object to be updated
    :param session: sqlalchemy scoped session
    :param paginate: Enable/disable pagination
    :param page_size: Number maximum elements showed in a page
    :param search_params: Query parameters
    :param path: endpoint
    :param collection: True if the type_ is of a collection, False for any other class
    :return: response containing a page of the objects of that particular type_

    Raises:
        ClassNotFound: If `type_` does not represent a valid/defined RDFClass.

    """
    database_search_params = copy.deepcopy(search_params)
    pagination_parameters = ["page", "pageIndex", "limit", "offset"]

    # remove pagination params before filtering in the database
    for param in database_search_params.copy():
        if param in pagination_parameters:
            database_search_params.pop(param)
    database_search_params = parse_search_params(database_search_params)
    filtered_instances = get_all_filtered_instances(
        session, database_search_params, type_, collection)
    collection_template = pagination(filtered_instances, path, type_, API_NAME,
                                     search_params, paginate, page_size)
    return collection_template


def get_single(type_: str, api_name: str, session: scoped_session,
               path: str = None) -> Dict[str, Any]:
    """Get instance of classes with single objects.
    :param type_: type of object to be updated
    :param api_name: api name specified while starting server
    :param session: sqlalchemy scoped session
    :param path: endpoint
    :return: response containing information about a single object

    Raises:
        ClassNotFound: If `type_` does not represt a valid/defined RDFClass.
        InstanceNotFound: If no Instance with type `type_` exists.

    """
    instance = get_single_response(session, type_)
    object_ = get(instance.id, type_, session=session, api_name=api_name, path=path)
    if path is not None:
        object_["@id"] = f"{get_host_domain()}/{api_name}/{path}"
    else:
        object_["@id"] = f"{get_host_domain()}/{api_name}/{type_}"
    return object_


def insert_single(object_: Dict[str, Any], session: scoped_session) -> Any:
    """Insert instance of classes with single objects.
    :param object_: object to be inserted
    :param session: sqlalchemy scoped session
    :return:

    Raises:
        ClassNotFound: If `type_` does not represt a valid/defined RDFClass.
        Instance: If an Instance of type `type_` already exists.

    """
    type_ = object_["@type"]
    database_class = get_database_class(type_)

    try:
        session.query(database_class).all()[-1]
    except (NoResultFound, IndexError, ValueError):
        return insert(object_, session=session)

    raise InstanceExists(type_)


def update_single(object_: Dict[str,
                                Any],
                  session: scoped_session,
                  api_name: str,
                  path: str = None) -> int:
    """Update instance of classes with single objects.
    :param object_: new object
    :param session: sqlalchemy scoped session
    :param api_name: api name specified while starting server
    :param path: endpoint
    :return: id of the updated object

    Raises:
        ClassNotFound: If `object['@type']` does not represt a valid/defined RDFClass.
        InstanceNotFound: If no Instance of the class exists.

    """
    type_ = object_["@type"]
    instance = get_single_response(session, type_)

    return update(
        id_=instance.id,
        type_=type_,
        object_=object_,
        session=session,
        api_name=api_name,
        path=path)


def delete_single(type_: str, session: scoped_session) -> None:
    """Delete instance of classes with single objects.
    :param type_: type of object to be deleted
    :param session: sqlalchemy scoped session
    :return: None

    Raises:
        ClassNotFound: If `type_` does not represt a valid/defined RDFClass.
        InstanceNotFound: If no Instance of the class exists.

    """
    instance = get_single_response(session, type_)

    return delete(instance.id, type_, session=session)


def get_member(collection_id: str, member_id: str, type_: str, api_name: str,
               session: scoped_session, path: str = None) -> Dict[str, str]:
    """Retrieve an Instance with given IDs from the database [GET].
    :param collection_id: id of the collection to be fetched
    :param member_id: id of the member to be fetched
    :param type_: type of object
    :param api_name: name of api specified while starting server
    :param session: sqlalchemy scoped session
    :param path: endpoint
    :return: response to the request


    Raises:
        ClassNotFound: If the `type_` is not a valid/defined RDFClass.
        InstanceNotFound: If no Instance of the 'type_` class if found.

    """
    query_info = {
        "@type": type_,
        "member_id": member_id,
        "collection_id": collection_id
    }

    object_template = get_collection_member(query_info, session)
    object_template["@id"] = f"{get_host_domain()}/{api_name}/{path}/{collection_id}"

    return object_template


def delete_member(collection_id: str, member_id: str, type_: str, session: scoped_session) -> None:
    """Delete an Instance and all its relations from DB given id [DELETE].
    :param collection_id: id of the collection
    :param member_id: id of member to be deleted
    :param type_: type of object to be deleted
    :param session: sqlalchemy scoped session

    Raises:
        ClassNotFound: If `type_` does not represent a valid/defined RDFClass.
        InstanceNotFound: If no instace of type `type_` with id `id_` exists.

    """
    query_info = {
        "@type": type_,
        "member_id": member_id,
        "collection_id": collection_id
    }
    delete_collection_member(query_info, session)


def insert_modification_record(method: str, resource_url: str,
                               session: scoped_session) -> int:
    """
    Insert a modification record into the database.
    :param method: HTTP method type of related operation.
    :param resource_url: URL of resource modified.
    :param session: sqlalchemy session.
    :return: ID of new modification record.
    """
    modification = Modification(method=method, resource_url=resource_url)
    session.add(modification)
    session.commit()
    return modification.job_id


def get_last_modification_job_id(session: scoped_session) -> str:
    """
    Get job id of most recent modification record stored in the db.
    :param session: sqlalchemy session
    :return: job id of recent modification.
    """
    last_modification = session.query(Modification).order_by(Modification.job_id.desc()).first()
    if last_modification is None:
        last_job_id = ""
    else:
        last_job_id = last_modification.job_id
    return last_job_id


def get_modification_table_diff(session: scoped_session,
                                agent_job_id: str = None) -> List[Dict[str, Any]]:
    """
    Get modification table difference.
    :param session: sqlalchemy session.
    :param agent_job_id: Job id from the client.
    :return: List of all modifications done after job with job_id = agent_job_id.
    """
    # If agent_job_id is not given then return all the elements.
    if agent_job_id is None:
        modifications = session.query(Modification).order_by(
            Modification.job_id.asc()).all()
    # If agent_job_id is given then return all records which are older
    # than the record with agent_job_id.
    else:
        try:
            record_for_agent_job_id = session.query(Modification).filter(
                Modification.job_id == agent_job_id).one()
        except NoResultFound:
            return []
        modifications = session.query(Modification).filter(
            Modification.job_id > record_for_agent_job_id.job_id).order_by(
            Modification.job_id.asc()).all()

    # Create response body
    list_of_modification_records = []
    for modification in modifications:
        modification_record = {
            "job_id": modification.job_id,
            "method": modification.method,
            "resource_url": modification.resource_url
        }
        list_of_modification_records.append(modification_record)
    return list_of_modification_records


def pagination(filtered_instances, path, type_, API_NAME,
               search_params, paginate, page_size):
    """Add pagination to the response and return the response
    :param filtered_instances: instances after filtered from the database query
    :param path: endpoint
    :param type_: type of object to be updated
    :param API_NAME: api name specified while starting server
    :param search_params: Query parameters
    :param paginate: Enable/disable pagination
    :param page_size: Number maximum elements showed in a page
    :return: response containing a page of the objects of that particular type_
    """
    collection_template = {
        "@id": f"{get_host_domain()}/{API_NAME}/{path}/",
        "@context": None,
        "@type": f"{path}",
        "members": []
    }  # type: Dict[str, Any]
    result_length = len(filtered_instances)
    try:
        # To paginate, calculate offset and page_limit values for pagination of search results
        page, page_size, offset = pre_process_pagination_parameters(search_params=search_params,
                                                                    paginate=paginate,
                                                                    page_size=page_size,
                                                                    result_length=result_length)
    except (IncompatibleParameters, PageNotFound, OffsetOutOfRange):
        raise
    current_page_size = page_size
    if result_length - offset < page_size:
        current_page_size = result_length - offset
    for i in range(offset, offset+current_page_size):
        object_template = {
            "@id": f"{get_host_domain()}/{API_NAME}/{type_}/{filtered_instances[i].id}",
            "@type": type_
        }
        collection_template["members"].append(object_template)

    # If pagination is disabled then stop and return the collection template
    collection_template["hydra:totalItems"] = result_length
    if paginate is False:
        return collection_template
    # Calculate last page number
    if result_length != 0 and result_length % page_size == 0:
        last = result_length // page_size
    else:
        last = result_length // page_size + 1
    if page < 1 or page > last:
        raise PageNotFound(str(page))
    recreated_iri = recreate_iri(API_NAME, path, search_params=search_params)
    # Decide which parameter to use to provide navigation
    if "offset" in search_params:
        paginate_param = "offset"
    elif "pageIndex" in search_params:
        paginate_param = "pageIndex"
    else:
        paginate_param = "page"
    attach_hydra_view(collection_template=collection_template, paginate_param=paginate_param,
                      result_length=result_length, iri=recreated_iri, page_size=page_size,
                      offset=offset, page=page, last=last)
    return collection_template


def item_exists(item_type, item_id, session):
    """Check if the instance with type 'item_type' and id 'item_id'
    exists in the database. Returns True if exists else False

    :param item_type: The @type of the instance
    :type item_type: str
    :param item_id: The id of the instance in the database
    :type item_id: str
    :param session: sqlalchemy scoped session
    """
    database_class = get_database_class(item_type)
    return session.query(database_class.id).filter_by(id=item_id).scalar() is not None



from typing import Union
from functools import wraps

from flask import jsonify, Response, request
from hydra_python_core.doc_writer import HydraError
from hydrus.utils import get_session, get_token, get_authentication

from hydrus.data.user import (create_nonce, check_authorization,
                              add_token, check_token)

from hydrus.helpers import set_response_headers, error_response


def token_response(token: str) -> Response:
    """
    Return succesful token generation object
    """
    message = {200: "User token generated"}
    response = set_response_headers(jsonify(message), status_code=200,
                                    headers=[{'X-Authorization': token}])
    return response


def failed_authentication(incorrect: bool) -> Response:
    """
    Return failed authentication object.
    """
    if not incorrect:
        message = {401: "Need credentials to authenticate"}
        realm = 'Basic realm="Login required"'
    else:
        message = {401: "Incorrect credentials"}
        realm = 'Basic realm="Incorrect credentials"'
    nonce = create_nonce(get_session())
    response = set_response_headers(jsonify(message), status_code=401,
                                    headers=[{'WWW-Authenticate': realm},
                                             {'X-Authentication': nonce}])
    return response


def verify_user() -> Union[Response, None]:
    """
    Verify the credentials of the user and assign token.
    """
    try:
        auth = check_authorization(request, get_session())
        if auth is False:
            return failed_authentication(True)
        elif get_token():
            token = add_token(request, get_session())
            return token_response(token)
    except Exception as e:
        error = e.get_HTTP()  # type: HydraError
        return error_response(error)
    return None


def check_authentication_response() -> Union[Response, None]:
    """
    Return the response as per the authentication requirements.
    """
    if get_authentication():
        if get_token():
            token = check_token(request, get_session())
            if not token:
                if request.authorization is None:
                    return failed_authentication(False)
                else:
                    return verify_user()
        elif request.authorization is None:
            return failed_authentication(False)
        else:
            return verify_user()
    else:
        return None


def authenticate(f):
    """
    Decorator for checking authentication of each request.
    Returns the authentication response if required, else
    continues with the request.
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        auth_response = check_authentication_response()
        if isinstance(auth_response, Response):
            return auth_response
        return f(*args, **kwargs)
    return wrapper



"""Models for Hydra Classes."""

import datetime
import uuid
from sqlite3 import Connection as SQLite3Connection
from typing import Any

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
    event,
)
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.sql import func
from hydra_python_core.doc_writer import DocUrl
# from hydrus.settings import DB_URL

engine = create_engine("sqlite:///database.db")

Base = declarative_base()  # type: Any

EXPIRY_TIME = 2700  # unit: seconds


# required for enforcing foreign key constraints through SQLite
@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection, connection_record):
    if isinstance(dbapi_connection, SQLite3Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()


class Resource:
    """
    Class for interacting with any resource(class or collection) in the ApiDoc
    """

    # TODO: Look for a better way to store the sql alchemy class for each resource
    all_database_classes = {}

    def __init__(self, resource):
        self._resource = resource

    @property
    def name(self):
        """Return the name of the resource from it's "@id"""
        # get the base url which will be used to split the @id
        # for eg, 'http://localhost:8080/serverapi/vocab#'
        expanded_base_url = DocUrl.doc_url
        return self._resource["@id"].split(expanded_base_url)[1]

    @property
    def resource(self):
        """Return the resource dict"""
        return self._resource

    @property
    def supported_properties(self):
        """
        Return all the properties in "supportedProperty" property
        of that resource.
        """
        return self._resource["supportedProperty"]

    def get_attr_dict(self):
        """
        Return the attribute dictionary necessary for
        creating that resource's table at runtime
        """
        # initialize the attribute dict with "__tablename__"
        # and a "id" column which will act as primary key
        # for instances of that table
        attr_dict = {
            "__tablename__": self.name,
            "id": Column(
                String,
                default=lambda: str(uuid.uuid4()),
                unique=True,
                primary_key=True,
            ),
        }
        for supported_property in self.supported_properties:
            title = supported_property["title"]
            property_ = supported_property["property"]
            expanded_base_url = DocUrl.doc_url
            if expanded_base_url in property_:
                # if expanded_base_url is in the property_, it implies that the property_
                # is pointing to another resource in the same ApiDoc, hence make
                # it a Foreign Key to that resource table
                foreign_table_name = property_.split(expanded_base_url)[1]
                attr_dict[title] = Resource.foreign_key_column(
                    foreign_table_name, title
                )
            else:
                attr_dict[title] = Column(String)

        if "manages" in self.resource:
            # if the class is a collection, add an extra column for
            # collection id
            attr_dict['collection_id'] = Column(
                String,
                default=lambda: str(uuid.uuid4()),
            )
            # if the class is a collection, add an extra column for
            # member @type
            attr_dict['member_type'] = Column(
                String,
            )
            # if the class is a collection, then a member id and collection id
            # should be a unique constraint altogether
            attr_dict['__table_args__'] = (
                UniqueConstraint(title, 'collection_id'),
            )
        return attr_dict

    @staticmethod
    def foreign_key_column(foreign_table_name, title):
        """
        Return a sqlalchemy column which will act as
        a foreign key to given tablename.
        :param foreign_table_name: The table name to which the foreign key
        relationship has to established
        :param title: The name of this foreign key column
        :return: A SQL-Alchemy column with correct foreign key attached
        """
        # title is to dereference the column name later
        return Column(
            String,
            ForeignKey(
                f"{foreign_table_name}.id",
                ondelete="CASCADE",
                onupdate="CASCADE",
                info={"column_name": title},
            ),
        )

    def make_db_table(self):
        """Generate the sqlalchemy table class for that resource"""
        self.table_class = type(self.name, (Base,), self.get_attr_dict())
        # add that class to dict for future lookups
        Resource.all_database_classes[self.name] = self.table_class


class Modification(Base):
    """Model for sync related state-changing modifications."""

    __tablename__ = "modifications"

    job_id = Column(Integer, primary_key=True, autoincrement=True)
    method = Column(String)
    resource_url = Column(String)


class User(Base):
    """Model for a user that stores the ID, paraphrase and a nonce."""

    __tablename__ = "user"

    id = Column(Integer, primary_key=True)
    paraphrase = Column(String)


class Token(Base):
    """Model for storing tokens for the users."""

    __tablename__ = "tokens"
    id = Column(
        String,
        default=lambda: str(uuid.uuid4()),
        unique=True,
        primary_key=True,
    )
    user_id = Column(Integer, ForeignKey("user.id"))
    timestamp = Column(DateTime, default=func.now())
    expiry = Column(
        "expiry",
        DateTime,
        default=datetime.datetime.utcnow() + datetime.timedelta(seconds=EXPIRY_TIME),
    )

    def is_valid(self):
        if self.expiry > datetime.datetime.utcnow():
            return True
        return False


class Nonce(Base):
    """Model for storing nonce for the users."""

    __tablename__ = "nonce"
    id = Column(String, primary_key=True)
    timestamp = Column(DateTime)


def create_database_tables(classes):
    for single_class in classes:
        resource = Resource(single_class)
        resource.make_db_table()


if __name__ == "__main__":
    print("Creating models....")
    Base.metadata.create_all(engine)
    print("Done")



"""Imports :
    flask.json.jsonify : Turns the JSON output into a Response object with the
    application/json mimetype Ref- http://flask.pocoo.org/docs/0.12/api
    flask.request : The request object used by default in Flask.
    Remembers the matched endpoint and view arguments.
    Ref - http://flask.pocoo.org/docs/0.12/api
    flask.abort : Raises an HTTPException for the given status code or WSGI
    application: Ref - http://flask.pocoo.org/docs/0.12/api
    flask_restful.Resource : Represents an abstract RESTful resource.
    Ref - http://flask-restful.readthedocs.io/en/latest/api.html
    hydrus.data.crud : Function/Class to perform basic CRUD operations for the server
    hydrus.auth.authenticate : Decorator for checking authentication of each request
    hydrus.utils.get_session : Gets the database session for the server
    hydrus.utils.get_doc : Function which gets the server API documentation
    hydrus.utils.get_api_name : Function which gets the server API name
    hydrus.utils.get_hydrus_server_url : Function the gets the server URL
    hydrus.utils.get_authentication : Function that checks whether API needs to be
    authenticated or not
    hydrus.utils.get_collections_and_parsed_classes: Function that returns
    all the collections and parsed classes in the API documentation
"""  # nopep8

import json

from flask import Response, jsonify, request, abort
from flask_restful import Resource

from hydrus.auth import authenticate
from hydrus.helpers import (
    set_response_headers,
    checkClassOp,
    checkEndpoint,
    check_writeable_props,
    get_context,
    get_fragments
)
from hydrus.utils import get_doc, get_collections_and_parsed_classes
from hydrus.itemhelpers import (
    items_get_check_support,
    items_post_check_support,
    items_put_check_support,
    items_delete_check_support,
    member_get_check_support,
    member_delete_check_support
)
from hydrus.item_collection_helpers import (
    item_collection_get_response,
    item_collection_put_response
)
from hydrus.items_helpers import (
    items_put_response,
    items_delete_response
)


class Index(Resource):
    """Class for the EntryPoint."""

    def get(self) -> Response:
        """Return main entrypoint for the api."""
        return set_response_headers(jsonify(get_doc().entrypoint.get()))


class Vocab(Resource):
    """Vocabulary for Hydra."""

    def get(self) -> Response:
        """Return the main hydra vocab or a fragment of the main hydra vocab."""
        try:
            resource = request.args.getlist('resource')[0]
            return set_response_headers(jsonify(get_fragments(resource)))
        except:
            return set_response_headers(jsonify(get_doc().generate()))


class Entrypoint(Resource):
    """Hydra EntryPoint."""

    def get(self) -> Response:
        """Return application main Entrypoint."""
        response = {"@context": get_doc().entrypoint.context.generate()}
        return set_response_headers(jsonify(response))


class Item(Resource):
    """Handles all operations(GET, POST, PATCH, DELETE) on Items
    (item can be anything depending upon the vocabulary)."""
    @authenticate
    def get(self, id_: str, path: str) -> Response:
        """
        GET object with id = id_ from the database.
        :param id_ : Item ID
        :param path : Path for Item ( Specified in APIDoc @id)
        :return : object with id=id_
        """
        id_ = str(id_)
        collections, parsed_classes = get_collections_and_parsed_classes()
        is_collection = False
        if path in parsed_classes:
            class_path = path
            class_type = parsed_classes[path]['class'].title
        if path in collections:
            item_class = collections[path]["collection"]
            class_type = item_class.name
            # Get path of the collection-class
            class_path = item_class.path
            is_collection = True
        if checkClassOp(class_path, "GET"):
            return items_get_check_support(id_, class_type, class_path, path, is_collection)
        abort(405)

    @authenticate
    def post(self, id_: str, path: str) -> Response:
        """
        Update object of type<path> at ID<id_> with new object_ using HTTP POST.
        :param id_ - ID of Item to be updated
        :param path - Path for Item type( Specified in APIDoc @id)
        """
        id_ = str(id_)
        collections, parsed_classes = get_collections_and_parsed_classes()
        is_collection = False
        if path in parsed_classes:
            class_path = path
        if path in collections:
            item_class = collections[path]["collection"]
            class_path = item_class.path
            is_collection = True
        object_ = json.loads(request.data.decode('utf-8'))
        if checkClassOp(class_path, "POST") and check_writeable_props(class_path, object_):
            return items_post_check_support(id_, object_, class_path, path, is_collection)
        abort(405)

    @authenticate
    def put(self, id_: str, path: str) -> Response:
        """
        Add new object_ optional <id_> parameter using HTTP PUT.
        :param id_ - ID of Item to be updated
        :param path - Path for Item type( Specified in APIDoc @id) to be updated
        """
        id_ = str(id_)
        collections, parsed_classes = get_collections_and_parsed_classes()
        is_collection = False
        if path in parsed_classes:
            class_path = path
        if path in collections:
            item_class = collections[path]["collection"]
            class_path = item_class.path
            is_collection = True
        if checkClassOp(class_path, "PUT"):
            return items_put_check_support(id_, class_path, path, is_collection)
        abort(405)

    @authenticate
    def delete(self, id_: str, path: str) -> Response:
        """
        Delete object with id=id_ from database.
        :param id_ - ID of Item to be deleted
        :param path - Path for Item type( Specified in APIDoc @id) to be deleted
        """
        id_ = str(id_)
        collections, parsed_classes = get_collections_and_parsed_classes()
        is_collection = False
        if path in parsed_classes:
            class_path = path
            class_type = parsed_classes[path]['class'].title
        if path in collections:
            item_class = collections[path]["collection"]
            class_type = item_class.name
            # Get path of the collection-class
            class_path = item_class.path
            is_collection = True
        if checkClassOp(class_path, "DELETE"):
            return items_delete_check_support(id_, class_type, path, is_collection)
        abort(405)


class ItemCollection(Resource):
    """Handle operation related to ItemCollection (a collection of items)."""
    @authenticate
    def get(self, path: str) -> Response:
        """
        Retrieve a collection of items from the database.
        :param path : Path of the Collection
        :return : collection of items
        """
        endpoint_ = checkEndpoint("GET", path)
        if not endpoint_['method']:
            # If endpoint and Get method not supported in the API
            abort(endpoint_['status'])
        return item_collection_get_response(path)

    @authenticate
    def put(self, path: str) -> Response:
        """
        Method executed for PUT requests.
        Used to add an item to a collection
        :param path - Path for Item type ( Specified in APIDoc @id)
        """
        endpoint_ = checkEndpoint("PUT", path)
        if not endpoint_['method']:
            # If endpoint and PUT method is not supported in the API
            abort(endpoint_['status'])
        return item_collection_put_response(path)


class ItemMember(Resource):
    """Handles operations(GET,DELETE) related to member of an Item(Collection).
    (Item should be hydra:Collection)"""
    @authenticate
    def get(self, id_: str, path: str, collection_id_: str) -> Response:
        """
        GET object with collection_id = collection_id_
        and member = id_ (member of a collection) from the database.
        :param id_ : Item ID (Member)
        :param collection_id_ : Item ID (Collection)
        :param path : Path for Item ( Specified in APIDoc @id)
        :return : object with member=id_ and collection_id=collection_id_
        """
        collection_id = str(collection_id_)
        member_id = str(id_)
        collections, parsed_classes = get_collections_and_parsed_classes()
        if path in parsed_classes:
            abort(405)
        if path in collections:
            item_class = collections[path]["collection"]
            class_type = item_class.name
            # Get path of the collection-class
            class_path = item_class.path
        if checkClassOp(class_path, "GET"):
            return member_get_check_support(collection_id, member_id, class_type, class_path, path)
        abort(405)

    @authenticate
    def delete(self, id_: str, path: str, collection_id_: str) -> Response:
        """
        Delete object with id=id_ from database.
        :param id_ - ID of Item to be deleted
        :param path - Path for Item type( Specified in APIDoc @id) to be deleted
        """
        collection_id = str(collection_id_)
        member_id = str(id_)
        collections, parsed_classes = get_collections_and_parsed_classes()
        if path in parsed_classes:
            abort(405)
        if path in collections:
            item_class = collections[path]["collection"]
            class_type = item_class.name
            # Get path of the collection-class
            class_path = item_class.path
        if checkClassOp(class_path, "DELETE"):
            return member_delete_check_support(collection_id, member_id, class_type, path)
        abort(405)


class Items(Resource):
    """Handles operations(PUT,DELETE) related to multiple objects.
    (Item should be hydra:Class)"""
    @authenticate
    def put(self, path, int_list="") -> Response:
        """
        To insert multiple objects into the database
        :param path: endpoint
        :param int_list: Optional String containing ',' separated ID's
        :return:
        """
        endpoint_ = checkEndpoint("PUT", path)
        if not endpoint_['method']:
            # If endpoint and PUT method is not supported in the API
            abort(endpoint_['status'])
        return items_put_response(path, int_list)

    @authenticate
    def delete(self, path, int_list):
        """
        To delete multiple objects
        :param path: endpoints
        :param int_list: Optional String containing ',' separated ID's
        :return:
        """
        return items_delete_response(path, int_list)


class Contexts(Resource):
    """Dynamically generated contexts."""

    def get(self, category: str) -> Response:
        """Return the context for the specified class."""
        return get_context(category)



from flask_socketio import Namespace, emit
from hydrus.data.crud import get_modification_table_diff, get_last_modification_job_id


class SyncNamespace(Namespace):

    def __init__(self, namespace, db_session):
        Namespace.__init__(self, namespace=namespace)
        self.db_session = db_session

    def on_connect(self):
        print('A client connected')
        emit('connect', {'last_job_id': get_last_modification_job_id(self.db_session)})

    def on_disconnect(self):
        print('A client disconnected')

    def on_reconnect(self):
        print('A client reconnected.')
        emit('connect', {'last_job_id': get_last_modification_job_id(self.db_session)})

    def on_get_modification_table_diff(self, data):
        """Get modification table diff and emit it to the client.
        :param data: Dict with 'agent_job_id' key.
        """
        if 'agent_job_id' in data:
            agent_job_id = data['agent_job_id']
            modification_table_diff = get_modification_table_diff(self.db_session,
                                                                  agent_job_id)
        else:
            modification_table_diff = get_modification_table_diff(self.db_session)
        emit('modification_table_diff', modification_table_diff)



"""
This file (test_app.py) contains the functional tests for checking if the
response format is proper. Run tests/unit/test_crud.py before running this.
"""

import json
import re
import uuid

import pytest
from hydra_python_core.doc_writer import HydraLink, DocUrl

from hydrus.tests.conftest import gen_dummy_object
from hydrus.utils import get_doc


# specify common fixture for all tests
@pytest.mark.usefixtures('init_db_for_app_tests')
class TestApp():
    def test_Index(self, test_app_client, constants):
        """Test for the Index."""
        HYDRUS_SERVER_URL = constants['HYDRUS_SERVER_URL']
        API_NAME = constants['API_NAME']
        response_get = test_app_client.get(f'/{API_NAME}')
        endpoints = json.loads(response_get.data.decode('utf-8'))
        response_post = test_app_client.post(f'/{API_NAME}', data=dict(foo='bar'))
        response_put = test_app_client.put(f'/{API_NAME}', data=dict(foo='bar'))
        response_delete = test_app_client.delete(f'/{API_NAME}')
        assert '@context' in endpoints
        assert endpoints['@id'] == f'{HYDRUS_SERVER_URL}{API_NAME}'
        assert endpoints['@type'] == 'EntryPoint'
        assert response_get.status_code == 200
        assert response_post.status_code == 405
        assert response_put.status_code == 405
        assert response_delete.status_code == 405

    def test_EntryPoint_context(self, test_app_client, constants):
        """Test for the EntryPoint context."""
        API_NAME = constants['API_NAME']
        response_get = test_app_client.get(f'/{API_NAME}/contexts/EntryPoint.jsonld')
        response_get_data = json.loads(response_get.data.decode('utf-8'))
        response_post = test_app_client.post(f'/{API_NAME}/contexts/EntryPoint.jsonld', data={})
        response_delete = test_app_client.delete(f'/{API_NAME}/contexts/EntryPoint.jsonld')
        assert response_get.status_code == 200
        assert '@context' in response_get_data
        assert response_post.status_code == 405
        assert response_delete.status_code == 405

    def test_Vocab(self, test_app_client, constants):
        """Test the vocab."""
        API_NAME = constants['API_NAME']
        HYDRUS_SERVER_URL = constants['HYDRUS_SERVER_URL']
        vocab_route = get_doc().doc_name
        response_get = test_app_client.get(f'/{API_NAME}/{vocab_route}#')
        response_get_data = json.loads(response_get.data.decode('utf-8'))

        assert '@context' in response_get_data
        assert response_get_data['@type'] == 'ApiDocumentation'
        assert response_get_data['@id'] == f'{HYDRUS_SERVER_URL}{API_NAME}/{vocab_route}'
        assert response_get.status_code == 200

        response_delete = test_app_client.delete(f'/{API_NAME}/{vocab_route}#')
        assert response_delete.status_code == 405

        response_put = test_app_client.put(
            f'/{API_NAME}/{vocab_route}#', data=json.dumps(dict(foo='bar')))
        assert response_put.status_code == 405

        response_post = test_app_client.post(f'/{API_NAME}/{vocab_route}#',
                                             data=json.dumps(dict(foo='bar')))
        assert response_post.status_code == 405

    def test_fragments(self, test_app_client, constants):
        """Test the fragments in vocab."""
        API_NAME = constants['API_NAME']
        HYDRUS_SERVER_URL = constants['HYDRUS_SERVER_URL']
        vocab_route = get_doc().doc_name
        response_get = test_app_client.get(f'/{API_NAME}/{vocab_route}#')
        response_get_data = json.loads(response_get.data.decode('utf-8'))

        for class_ in response_get_data['supportedClass']:
            resource_id = class_['@id']
            if '#' not in resource_id:
                resource_name = resource_id.split('?resource=')[-1]
                vocab_uri = f'{HYDRUS_SERVER_URL}{API_NAME}/{vocab_route}'
                param_string = f'?resource={resource_name}'
                response_fragment_get = test_app_client.get(f'{vocab_uri}{param_string}')
                response_fragment_get_data = json.loads(response_fragment_get.data.decode('utf-8'))

                assert response_fragment_get.status_code == 200
                assert '@context' in response_fragment_get_data
                assert response_fragment_get_data['supportedClass'][0]['@id'] == resource_id
                assert len(response_fragment_get_data['supportedClass']) == 1

    def test_Collections_GET(self, test_app_client, constants, doc, init_db_for_app_tests):
        """Test GET on collection endpoints."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            response_get = test_app_client.get(endpoint['@id'])
            assert response_get.status_code == 200
            response_get_data = json.loads(response_get.data.decode('utf-8'))
            assert '@context' in response_get_data
            assert '@id' in response_get_data
            assert '@type' in response_get_data
            assert 'members' in response_get_data
            # Check the item URI has the valid format, so it can be dereferenced
            if len(response_get_data['members']) > 0:
                for item in response_get_data['members']:
                    class_type = item['@type']
                    if class_type in doc.parsed_classes:
                        class_ = doc.parsed_classes[class_type]['class']
                        class_methods = [
                            x.method for x in class_.supportedOperation]
                        if 'GET' in class_methods:
                            item_response = test_app_client.get(
                                response_get_data['members'][0]['@id'])
                            assert item_response.status_code == 200

    def test_pagination(self, test_app_client, constants, doc):
        """Test basic pagination"""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            response_get = test_app_client.get(endpoint['@id'])
            assert response_get.status_code == 200
            response_get_data = json.loads(
                response_get.data.decode('utf-8'))
            assert 'hydra:view' in response_get_data
            assert 'hydra:first' in response_get_data['hydra:view']
            assert 'hydra:last' in response_get_data['hydra:view']
            if 'hydra:next' in response_get_data['hydra:view']:
                response_next = test_app_client.get(
                    response_get_data['hydra:view']['hydra:next'])
                assert response_next.status_code == 200
                response_next_data = json.loads(response_next.data.decode('utf-8'))
                assert 'hydra:previous' in response_next_data['hydra:view']
            break

    def test_Collections_PUT(self, test_app_client, constants, doc):
        """Test insert data to the collection."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            if 'PUT' in collection_methods:
                dummy_object = gen_dummy_object(collection.name, doc)
                good_response_put = test_app_client.put(endpoint['@id'],
                                                        data=json.dumps(dummy_object))
                assert good_response_put.status_code == 201
                assert good_response_put.json["iri"] == good_response_put.location

    def test_Collections_constraint_PUT(self, test_app_client, constants, doc):
        """Test collection constraints by inserting same object twice."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            if 'PUT' in collection_methods:
                dummy_object = gen_dummy_object(collection.name, doc)
                good_response_put = test_app_client.put(endpoint['@id'],
                                                        data=json.dumps(dummy_object))
                assert good_response_put.status_code == 201
                assert good_response_put.json["iri"] == good_response_put.location
                collection_id = good_response_put.location
                bad_response_put = test_app_client.put(collection_id,
                                                       data=json.dumps(dummy_object))
                assert bad_response_put.status_code == 400

    def test_collection_object_GET(self, test_app_client, constants, doc):
        """Test GET of a given collection object using ID."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            if 'PUT' in collection_methods:
                dummy_object = gen_dummy_object(collection.name, doc)
                initial_put_response = test_app_client.put(
                    endpoint['@id'], data=json.dumps(dummy_object))
                assert initial_put_response.status_code == 201
                response = json.loads(initial_put_response.data.decode('utf-8'))
                regex = r'(.*)ID (.{36})* (.*)'
                matchObj = re.match(regex, response['description'])
                assert matchObj is not None
                id_ = matchObj.group(2)
                if 'GET' in collection_methods:
                    get_response = test_app_client.get(f'{endpoint["@id"]}/{id_}')
                    assert get_response.status_code == 200

    def test_collection_object_PUT(self, test_app_client, constants, doc):
        """Test PUT of a given collection object using ID."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            if 'PUT' in collection_methods:
                dummy_object_1 = gen_dummy_object(collection.name, doc)
                initial_put_response = test_app_client.put(
                    endpoint['@id'], data=json.dumps(dummy_object_1))
                assert initial_put_response.status_code == 201
                collection_id = initial_put_response.location
                dummy_object_2 = gen_dummy_object(collection.name, doc)
                second_put_response = test_app_client.put(
                    collection_id, data=json.dumps(dummy_object_2))
                assert second_put_response.status_code == 201
                assert second_put_response.location == collection_id
                assert second_put_response.json["iri"] == collection_id

    def test_collection_object_POST(self, test_app_client, constants, doc, socketio):
        """Test POST of a given collection object using ID."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            dummy_object = gen_dummy_object(collection.name, doc)
            initial_put_response = test_app_client.put(
                endpoint['@id'], data=json.dumps(dummy_object))
            assert initial_put_response.status_code == 201
            assert initial_put_response.json["iri"] == initial_put_response.location
            response = json.loads(initial_put_response.data.decode('utf-8'))
            regex = r'(.*)ID (.{36})* (.*)'
            matchObj = re.match(regex, response['description'])
            assert matchObj is not None
            id_ = matchObj.group(2)
            if 'POST' in collection_methods:
                # members attribute should be writeable for POSTs
                if collection.supportedProperty[0].write:
                    dummy_object = gen_dummy_object(collection.name, doc)
                    post_replace_response = test_app_client.post(f'{endpoint["@id"]}/{id_}',
                                                                 data=json.dumps(dummy_object))
                    assert post_replace_response.status_code == 200

    def test_collection_object_DELETE(self, test_app_client, constants, doc):
        """Test DELETE of a given collection object using ID."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(
                endpoint["@id"].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [
                x.method for x in collection.supportedOperation]
            dummy_object = gen_dummy_object(collection.name, doc)
            initial_put_response = test_app_client.put(endpoint["@id"],
                                                       data=json.dumps(dummy_object))
            assert initial_put_response.status_code == 201
            response = json.loads(initial_put_response.data.decode('utf-8'))
            regex = r'(.*)ID (.{36})* (.*)'
            matchObj = re.match(regex, response['description'])
            assert matchObj is not None
            id_ = matchObj.group(2)
            if 'DELETE' in collection_methods:
                delete_response = test_app_client.delete(
                    f'{endpoint["@id"]}/{id_}')
                assert delete_response.status_code == 200

    def test_object_PUT_at_id(self, test_app_client, constants, doc):
        """Create object in collection using PUT at specific ID."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(
                endpoint["@id"].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [
                x.method for x in collection.supportedOperation]
            dummy_object = gen_dummy_object(collection.name, doc)
            if 'PUT' in collection_methods:
                dummy_object = gen_dummy_object(collection.name, doc)
                put_response = test_app_client.put(f'{endpoint["@id"]}/{uuid.uuid4()}',
                                                   data=json.dumps(dummy_object))
                assert put_response.status_code == 201
                assert put_response.json["iri"] == put_response.location

    def test_object_PUT_at_ids(self, test_app_client, constants, doc):
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    data_ = {'data': list()}
                    objects = list()
                    ids = ''
                    for index in range(3):
                        objects.append(gen_dummy_object(class_.title, doc))
                        ids = f'{uuid.uuid4()},'
                    data_['data'] = objects
                    if 'PUT' in class_methods:
                        put_response = test_app_client.put(f'{endpoints[endpoint]}/add/{ids}',
                                                           data=json.dumps(data_))
                        assert put_response.status_code == 201
                        assert isinstance(put_response.json['iri'], list)

    def test_endpointClass_PUT(self, test_app_client, constants, doc):
        """Check non collection Class PUT."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(
                    endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [
                        x.method for x in class_.supportedOperation]
                    if 'PUT' in class_methods:
                        dummy_object = gen_dummy_object(class_.title, doc)
                        put_response = test_app_client.put(endpoints[endpoint],
                                                           data=json.dumps(dummy_object))
                        assert put_response.status_code == 201
                        assert put_response.json["iri"] == put_response.location

    def test_endpointClass_POST(self, test_app_client, constants, doc):
        """Check non collection Class POST."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    if 'PUT' in class_methods:
                        # first insert a object which we will update later
                        dummy_object = gen_dummy_object(class_.title, doc)
                        initial_put_response = test_app_client.put(endpoints[endpoint],
                                                                   data=json.dumps(dummy_object))
                        response = json.loads(
                            initial_put_response.data.decode('utf-8'))
                        regex = r'(.*)ID (.{36})* (.*)'
                        matchObj = re.match(regex, response['description'])
                        id_ = matchObj.group(2)
                        if 'POST' in class_methods:
                            dummy_object = gen_dummy_object(class_.title, doc)
                            post_response = test_app_client.post(f'{endpoints[endpoint]}/{id_}',
                                                                 data=json.dumps(dummy_object))
                            assert post_response.status_code == 200

    def test_endpointClass_DELETE(self, test_app_client, constants, doc):
        """Check non collection Class DELETE."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(
                    endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    if 'PUT' in class_methods:
                        # first insert a object which we will update later
                        dummy_object = gen_dummy_object(class_.title, doc)
                        initial_put_response = test_app_client.put(endpoints[endpoint],
                                                                   data=json.dumps(dummy_object))
                        response = json.loads(
                            initial_put_response.data.decode('utf-8'))
                        regex = r'(.*)ID (.{36})* (.*)'
                        matchObj = re.match(regex, response['description'])
                        id_ = matchObj.group(2)
                        if 'DELETE' in class_methods:
                            delete_response = test_app_client.delete(
                                f'{endpoints[endpoint]}/{id_}')
                            assert delete_response.status_code == 200

    def test_endpointClass_GET(self, test_app_client, constants, doc):
        """Check non collection Class GET."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    if 'GET' in class_methods:
                        response_get = test_app_client.get(endpoints[endpoint])
                        assert response_get.status_code == 405

    def test_Collections_member_GET(self, test_app_client, constants, doc):
        """Test endpoint to get member from a collection"""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            if 'PUT' in collection_methods:
                dummy_object = gen_dummy_object(collection.name, doc)
                good_response_put = test_app_client.put(endpoint['@id'],
                                                        data=json.dumps(dummy_object))
                assert good_response_put.status_code == 201
                collection_endpoint = good_response_put.location
                if 'GET' in collection_methods:
                    for member in dummy_object['members']:
                        member_id = member['@id'].split('/')[-1]
                        get_response = test_app_client.get(f'{collection_endpoint}/{member_id}')
                        assert get_response.status_code == 200

    def test_Collections_member_DELETE(self, test_app_client, constants, doc):
        """Test endpoint to delete member from a collection."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint['@id'].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            collection_methods = [x.method for x in collection.supportedOperation]
            if 'PUT' in collection_methods:
                dummy_object = gen_dummy_object(collection.name, doc)
                good_response_put = test_app_client.put(endpoint['@id'],
                                                        data=json.dumps(dummy_object))
                assert good_response_put.status_code == 201
                collection_endpoint = good_response_put.location
                if 'DELETE' in collection_methods:
                    for member in dummy_object['members']:
                        member_id = member['@id'].split('/')[-1]
                        full_endpoint = f'{collection_endpoint}/{member_id}'
                        delete_response = test_app_client.delete(full_endpoint)
                        assert delete_response.status_code == 200

    def test_IriTemplate(self, test_app_client, constants, doc):
        """Test structure of IriTemplates attached to parsed classes"""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        expanded_base_url = DocUrl.doc_url
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint["@id"].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            class_name = collection.manages["object"].split(expanded_base_url)[1]
            response_get = test_app_client.get(endpoint["@id"])
            assert response_get.status_code == 200
            response_get_data = json.loads(response_get.data.decode('utf-8'))
            assert 'search' in response_get_data
            assert 'hydra:mapping' in response_get_data['search']
            class_ = doc.parsed_classes[class_name]['class']
            class_props = [x.prop for x in class_.supportedProperty]
            for mapping in response_get_data['search']['hydra:mapping']:
                prop = mapping['hydra:property']
                prop_name = mapping['hydra:variable']
                is_valid_class_prop = prop not in ['limit', 'offset', 'pageIndex']
                # check if IRI property is for searching through a nested_class
                # and not this class_
                is_nested_class_prop = "[" in prop_name and "]" in prop_name
                if is_valid_class_prop and not is_nested_class_prop:
                    assert prop in class_props

    def test_client_controlled_pagination(self, test_app_client, constants, doc):
        """Test pagination controlled by test_app_client with help of pageIndex,
        offset and limit parameters."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            response_get = test_app_client.get(endpoint["@id"])
            assert response_get.status_code == 200
            response_get_data = json.loads(response_get.data.decode('utf-8'))
            assert 'search' in response_get_data
            assert 'hydra:mapping' in response_get_data['search']
            # Test with pageIndex and limit
            params = {'pageIndex': 1, 'limit': 2}
            response_for_page_param = test_app_client.get(
                endpoint["@id"], query_string=params)
            assert response_for_page_param.status_code == 200
            response_for_page_param_data = json.loads(
                response_for_page_param.data.decode('utf-8'))
            assert 'hydra:first' in response_for_page_param_data['hydra:view']
            assert 'hydra:last' in response_for_page_param_data['hydra:view']
            if 'hydra:next' in response_for_page_param_data['hydra:view']:
                hydra_next = response_for_page_param_data['hydra:view']['hydra:next']
                assert 'pageIndex=2' in hydra_next
                next_response = test_app_client.get(
                    response_for_page_param_data['hydra:view']['hydra:next'])
                assert next_response.status_code == 200
                next_response_data = json.loads(
                    next_response.data.decode('utf-8'))
                assert 'hydra:previous' in next_response_data['hydra:view']
                data = next_response_data['hydra:view']['hydra:previous']
                assert 'pageIndex=1' in data
                # Test with offset and limit
                params = {'offset': 1, 'limit': 2}
                response_for_offset_param = test_app_client.get(endpoint["@id"],
                                                                query_string=params)
                assert response_for_offset_param.status_code == 200
                response_for_offset_param_data = json.loads(
                    response_for_offset_param.data.decode('utf-8'))
                data = response_for_offset_param_data['hydra:view']
                assert 'hydra:first' in data
                assert 'hydra:last' in data
                if 'hydra:next' in data:
                    assert 'offset=3' in data['hydra:next']
                    next_response = test_app_client.get(data['hydra:next'])
                    assert next_response.status_code == 200
                    next_response_data = json.loads(next_response.data.decode('utf-8'))
                    assert 'hydra:previous' in next_response_data['hydra:view']
                    assert 'offset=1' in next_response_data['hydra:view']['hydra:previous']

    def test_GET_for_nested_class(self, test_app_client, constants, doc):
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            collection_name = '/'.join(endpoint["@id"].split(f'/{API_NAME}/')[1:])
            collection = doc.collections[collection_name]['collection']
            class_methods = [x.method for x in collection.supportedOperation]
            if 'GET' in class_methods:
                response_get = test_app_client.get(endpoint["@id"])
                assert response_get.status_code == 200
                instance = response_get.json['members'][0]['@id']
                instance_type = instance.split('/')[-2]
                instance_class = doc.parsed_classes[instance_type]['class']
                instance_methods = [x.method for x in instance_class.supportedOperation]
                if 'GET' in instance_methods:
                    response_get_data = test_app_client.get(instance).json
                    assert '@context' in response_get_data
                    assert '@id' in response_get_data
                    assert '@type' in response_get_data
                    class_props = [x for x in collection.supportedProperty]
                    expanded_base_url = DocUrl.doc_url
                    for prop_name in class_props:
                        if not isinstance(prop_name.prop, HydraLink):
                            if expanded_base_url in prop_name.prop:
                                assert '@type' in response_get_data[prop_name.title]

    def test_required_props(self, test_app_client, constants, doc):
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    if 'PUT' in class_methods:
                        dummy_object = gen_dummy_object(class_.title, doc)
                        required_prop = ''
                        for prop in class_.supportedProperty:
                            if prop.required:
                                required_prop = prop.title
                                break
                        if required_prop:
                            del dummy_object[required_prop]
                            put_response = test_app_client.put(
                                endpoints[endpoint], data=json.dumps(dummy_object))
                            assert put_response.status_code == 400

    def test_writeable_props(self, test_app_client, constants, doc):
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    if 'PUT' in class_methods:
                        # first insert a object which we will update later
                        dummy_object = gen_dummy_object(class_.title, doc)
                        initial_put_response = test_app_client.put(endpoints[endpoint],
                                                                   data=json.dumps(dummy_object))
                        response = json.loads(initial_put_response.data.decode('utf-8'))
                        regex = r'(.*)ID (.{36})* (.*)'
                        matchObj = re.match(regex, response['description'])
                        id_ = matchObj.group(2)
                        if 'POST' in class_methods:
                            dummy_object = gen_dummy_object(class_.title, doc)
                            # Test for writeable properties
                            post_response = test_app_client.post(
                                f'{endpoints[endpoint]}/{id_}', data=json.dumps(dummy_object))
                            assert post_response.status_code == 200
                            # Test for properties with writeable=False
                            non_writeable_prop = ''
                            for prop in class_.supportedProperty:
                                if prop.write is False:
                                    non_writeable_prop = prop.title
                                    break
                            if non_writeable_prop != '':
                                dummy_object[non_writeable_prop] = 'xyz'
                                post_response = test_app_client.post(
                                    endpoints[endpoint], data=json.dumps(dummy_object))
                                assert post_response.status_code == 405

    def test_readable_props(self, test_app_client, constants, doc):
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(
                    endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [
                        x.method for x in class_.supportedOperation]
                    if 'GET' in class_methods:
                        not_readable_prop = ''
                        for prop in class_.supportedProperty:
                            if prop.read is False:
                                not_readable_prop = prop.title
                                break
                        if not_readable_prop:
                            get_response = test_app_client.get(
                                endpoints[endpoint])
                            get_response_data = json.loads(
                                get_response.data.decode('utf-8'))
                            assert not_readable_prop not in get_response_data

    def test_bad_objects(self, test_app_client, constants, doc):
        """Checks if bad objects are added or not."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(
                    endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name in doc.parsed_classes:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [
                        x.method for x in class_.supportedOperation]
                    if 'PUT' in class_methods:
                        bad_response_put = test_app_client.put(
                            endpoints[endpoint],
                            data=json.dumps(dict(foo='bar')))
                        assert bad_response_put.status_code == 400

    def test_bad_requests(self, test_app_client, constants, doc):
        """Checks if bad requests are handled or not."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                if class_name not in doc.collections:
                    class_ = doc.parsed_classes[class_name]['class']
                    class_methods = [x.method for x in class_.supportedOperation]
                    dummy_object = gen_dummy_object(class_.title, doc)
                    if 'PUT' in class_methods:
                        initial_put_response = test_app_client.put(
                            endpoints[endpoint], data=json.dumps(dummy_object))
                        assert initial_put_response.status_code == 201
                        response = json.loads(initial_put_response.data.decode('utf-8'))
                        regex = r'(.*)ID (.{36})* (.*)'
                        matchObj = re.match(regex, response['description'])
                        assert matchObj is not None
                        id_ = matchObj.group(2)
                        if 'POST' not in class_methods:
                            dummy_object = gen_dummy_object(class_.title, doc)
                            post_replace_response = test_app_client.post(
                                f'{endpoints[endpoint]}/{id_}', data=json.dumps(dummy_object))
                            assert post_replace_response.status_code == 405
                        if 'DELETE' not in class_methods:
                            delete_response = test_app_client.delete(
                                f'{endpoints[endpoint]}/{id_}')
                            assert delete_response.status_code == 405

    def test_Endpoints_Contexts(self, test_app_client, constants, doc):
        """Test all endpoints contexts are generated properly."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get(f'/{API_NAME}')
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints['collections']:
            response_get = test_app_client.get(endpoints['@id'])
            assert response_get.status_code == 200
            context = json.loads(response_get.data.decode('utf-8'))['@context']
            response_context = test_app_client.get(context)
            response_context_data = json.loads(
                response_context.data.decode('utf-8'))
            assert response_context.status_code == 200
            assert '@context' in response_context_data



"""
Script to generate the tables in hydrus database based on
resources in the provided API Doc.
"""
import copy
import uuid
from hydrus.data.db_models import Resource
from hydrus.data.exceptions import (
    ClassNotFound,
    InstanceExists,
    InstanceNotFound,
    InvalidSearchParameter,
    PropertyNotFound,
    PropertyNotGiven,
    MemberInstanceNotFound
)
from sqlalchemy import exists
from sqlalchemy.exc import InvalidRequestError, IntegrityError
from sqlalchemy.orm.scoping import scoped_session
from sqlalchemy.orm.exc import NoResultFound
from typing import Dict, Any


def get_type(object_: Dict[str, Any]) -> str:
    """
    Return the @type of that given object.
    :param object_: Dict containing object properties
    :return: The @type of that object
    """
    try:
        type_ = object_["@type"]
    except TypeError:
        raise ClassNotFound(object_)
    return type_


def get_database_class(type_: str):
    """
    Get the sqlalchemy class object from given classname
    :param type_: The @type of a object
    :return: The SQL-Alchemy database class object for that type_
    """
    database_class = Resource.all_database_classes.get(type_, None)
    if database_class is None:
        raise ClassNotFound(type_)
    return database_class


def insert_object(object_: Dict[str, Any], session: scoped_session,
                  collection: bool = False) -> str:
    """
    Insert the object in the database
    :param object_: Dict containing object properties
    :param session: sqlalchemy session
    :return: The ID of the inserted object
    """
    type_ = get_type(object_)
    database_class = get_database_class(type_)
    id_ = object_.get("id", None)
    if collection:
        # if type_ is of a collection class
        members = object_['members']
        collection_id = id_ if id_ else str(uuid.uuid4())
        for member in members:
            # add all the members of that collection
            inserted_object = database_class(members=member['id_'],
                                             collection_id=collection_id,
                                             member_type=member['@type'],
                                             )
            try:
                session.add(inserted_object)
                session.commit()
            except InvalidRequestError:
                session.rollback()
            except IntegrityError:
                session.rollback()
                member_type_ = member['@type']
                member_id_ = member['id_']
                raise InstanceExists(member_type_, member_id_)
        return collection_id
    else:
        # when type_ is of a non-collection class
        if (
            id_ is not None and
            session.query(exists().where(database_class.id == id_)).scalar()
        ):
            raise InstanceExists(type_, id_)
        foreign_keys = database_class.__table__.foreign_keys
        for fk in foreign_keys:
            # the name of the column through which this foreign key relationship
            # is being established
            fk_column = fk.info["column_name"]
            try:
                fk_object = object_[fk_column]
            except KeyError as e:
                wrong_property = e.args[0]
                raise PropertyNotGiven(type_=wrong_property)
            # insert the foreign key object
            fk_object_id = insert_object(fk_object, session)
            # put the id of the foreign instance in this table's column
            object_[fk_column] = fk_object_id
        try:
            # remove the @type from object before using the object to make a
            # instance of it using sqlalchemy class
            object_.pop("@type")
            inserted_object = database_class(**object_)
        except TypeError as e:
            # extract the wrong property name from TypeError object
            wrong_property = e.args[0].split("'")[1]
            raise PropertyNotFound(type_=wrong_property)
        try:
            session.add(inserted_object)
            session.commit()
        except InvalidRequestError:
            session.rollback()

        return inserted_object.id


def get_object(
    query_info: Dict[str, str], session: scoped_session, collection: bool = False
) -> Dict[str, str]:
    """
    Get the object from the database
    :param query_info: Dict containing the id and @type of object that has to retrieved
    :param session: sqlalchemy session
    :param collection: True if the type_ is of a collection, False for any other class
    :return: dict of object with its properties
    """
    type_ = query_info["@type"]
    id_ = query_info["id_"]
    database_class = get_database_class(type_)
    if collection:
        objects = (
            session.query(database_class.members, database_class.member_type)
            .filter(database_class.collection_id == id_)
            .all()
        )
        if len(objects) == 0:
            raise InstanceNotFound(type_=type_, id_=id_)
        object_template = {}
        object_template["@type"] = query_info["@type"]
        object_template["members"] = objects
        return object_template
    else:
        try:
            object_ = (
                session.query(database_class)
                .filter(database_class.id == id_)
                .one()
            ).__dict__
        except NoResultFound:
            raise InstanceNotFound(type_=type_, id_=id_)
        # Remove the unnecessary keys from the object retrieved from database
        object_template = copy.deepcopy(object_)
        object_template.pop("_sa_instance_state")
        object_template.pop("id")
        object_template["@type"] = query_info["@type"]
        return object_template


def delete_object(query_info: Dict[str, str], session: scoped_session,
                  collection: bool = False) -> None:
    """
    Delete the object from the database
    :param query_info: Dict containing the id and @type of object that has to retrieved
    :param session: sqlalchemy session
    :param collection: True if the type_ is of a collection, False for any other class
    """
    type_ = query_info["@type"]
    id_ = query_info["id_"]
    database_class = get_database_class(type_)
    if collection:
        try:
            objects = session.query(database_class).filter_by(collection_id=id_).delete()
        except NoResultFound:
            raise InstanceNotFound(type_=type_, id_=id_)
        try:
            session.commit()
        except InvalidRequestError:
            session.rollback()
        return id_
    else:
        try:
            object_ = (
                session.query(database_class)
                .filter(database_class.id == id_)
                .one()
            )
        except NoResultFound:
            raise InstanceNotFound(type_=type_, id_=id_)
        session.delete(object_)
        try:
            session.commit()
        except InvalidRequestError:
            session.rollback()


def update_object(
    object_: Dict[str, Any],
    query_info: Dict[str, str],
    session: scoped_session,
    collection: bool = False
) -> str:
    """
    Update the object from the database
    :param object_: Dict containing updated object properties
    :param query_info: Dict containing the id and @type of object that has to retrieved
    :param session: sqlalchemy session
    :return: the ID of the updated object
    """
    # Keep the object as fail safe
    old_object = get_object(query_info, session, collection)
    # Delete the old object
    delete_object(query_info, session, collection)
    id_ = query_info["id_"]
    # Try inserting new object
    try:
        object_["id"] = id_
        d = insert_object(object_, session, collection)
    except Exception as e:
        # Put old object back
        old_object["id"] = id_
        d = insert_object(old_object, session, collection)
        raise e
    return id_


def get_collection_member(query_info: Dict[str, str], session: scoped_session) -> Dict[str, str]:
    """
    Get member from a collection
    :param query_info: Dict containing the ids and @type of object that has to retrieved
    :param session: sqlalchemy session
    :return: dict of object with its properties
    """
    type_ = query_info["@type"]
    collection_id = query_info["collection_id"]
    member_id = query_info["member_id"]
    database_class = get_database_class(type_)
    objects = (
        session.query(database_class.members, database_class.member_type)
        .filter(database_class.collection_id == collection_id, database_class.members == member_id)
        .all()
    )
    if len(objects) == 0:
        raise MemberInstanceNotFound(type_=type_, collection_id_=collection_id,
                                     member_id_=member_id)
    object_template = {}
    object_template["@type"] = query_info["@type"]
    object_template["members"] = objects
    return object_template


def delete_collection_member(query_info: Dict[str, str], session: scoped_session) -> None:
    """
    Delete the object from the collection in database
    :param query_info: Dict containing the id and @type of object that has to retrieved
    :param session: sqlalchemy session
    """
    type_ = query_info["@type"]
    member_id = query_info["member_id"]
    collection_id = query_info["collection_id"]
    database_class = get_database_class(type_)
    objects = (
        session.query(database_class)
        .filter(database_class.collection_id == collection_id,
                database_class.members == member_id).delete()
    )
    if objects == 0:
        raise MemberInstanceNotFound(type_=type_, collection_id_=collection_id,
                                     member_id_=member_id)
    try:
        session.commit()
    except InvalidRequestError:
        session.rollback()
    return collection_id


def get_all_filtered_instances(
    session: scoped_session, search_params: Dict[str, Any], type_: str, collection: bool = False
):
    """Get all the filtered instances of from the database
    based on given query parameters.
    :param session: sqlalchemy scoped session
    :param search_params: Query parameters
    :param type_: @type of object to be deleted
    :param collection: True if the type_ is of a collection, False for any other class
    :return: filtered instances
    """
    database_class = get_database_class(type_)
    if collection:
        query = session.query(database_class.collection_id.label('id')).distinct()
    else:
        query = session.query(database_class)
        for param, param_value in search_params.items():
            # nested param
            if type(param_value) is dict:
                foreign_keys = database_class.__table__.foreign_keys
                for fk in foreign_keys:
                    if fk.info['column_name'] == param:
                        fk_table_name = fk.column.table.name
                        continue
                nested_param_db_class = get_database_class(fk_table_name)
                # build query
                for attr, value in param_value.items():
                    query = query.join(nested_param_db_class)
                    try:
                        query = query.filter(getattr(nested_param_db_class, attr) == value)
                    except AttributeError:
                        raise InvalidSearchParameter(f'{param}[{attr}]')
            else:
                value = search_params[param]
                try:
                    query = query.filter(getattr(database_class, param) == value)
                except AttributeError:
                    raise InvalidSearchParameter(f'{param}')

    filtered_instances = query.all()
    return filtered_instances


def get_single_response(session: scoped_session, type_: str):
    """
    Get instance of classes with single objects.
    :param session: sqlalchemy scoped session
    :param type_: @type of object to be deleted
    :return: instance
    """
    database_class = get_database_class(type_)
    try:
        instance = session.query(database_class).all()[-1]
    except (NoResultFound, IndexError, ValueError):
        raise InstanceNotFound(type_)

    return instance



"""
This file (test_socket.py) contains the functional tests for checking
socket events and operations
"""
import json

import pytest

import hydrus.data.crud as crud
from hydrus.tests.conftest import gen_dummy_object


@pytest.mark.usefixtures('init_db_for_socket_tests')
class TestSocket:
    def test_connect(self, socketio, app, session):
        """Test connect event."""
        socket_client = socketio.test_client(app, namespace='/sync')
        data = socket_client.get_received('/sync')
        assert len(data) > 0
        event = data[0]
        assert event['name'] == 'connect'
        last_job_id = crud.get_last_modification_job_id(session)
        assert event['args'][0]['last_job_id'] == last_job_id
        socket_client.disconnect(namespace='/sync')

    def test_reconnect(self, socketio, app, session):
        """Test reconnect event."""
        socket_client = socketio.test_client(app, namespace='/sync')
        # Flush data of first connect event
        socket_client.get_received('/sync')
        # Client reconnects by emitting 'reconnect' event.
        socket_client.emit('reconnect', namespace='/sync')
        # Get update received on reconnecting to the server
        data = socket_client.get_received('/sync')
        assert len(data) > 0
        # Extract the event information
        event = data[0]
        assert event['name'] == 'connect'
        last_job_id = crud.get_last_modification_job_id(session)
        # Check last job id with last_job_id received by client in the update.
        assert event['args'][0]['last_job_id'] == last_job_id
        socket_client.disconnect(namespace='/sync')

    def test_modification_table_diff(self, socketio_client, session):
        """Test 'modification-table-diff' events."""
        # Flush old received data at socket client
        socketio_client.get_received('/sync')
        # Set last_job_id as the agent_job_id
        agent_job_id = crud.get_last_modification_job_id(session)
        # Add an extra modification record newer than the agent_job_id
        new_latest_job_id = crud.insert_modification_record(method='POST',
                                                            resource_url='', session=session)
        socketio_client.emit('get_modification_table_diff',
                             {'agent_job_id': agent_job_id}, namespace='/sync')
        data = socketio_client.get_received('/sync')
        assert len(data) > 0
        event = data[0]
        assert event['name'] == 'modification_table_diff'
        # Check received event contains data of newly added modification record.
        assert event['args'][0][0]['method'] == 'POST'
        assert event['args'][0][0]['resource_url'] == ''
        assert event['args'][0][0]['job_id'] == new_latest_job_id

    def test_socketio_POST_updates(self, socketio_client, test_app_client, constants, doc):
        """Test 'update' event emitted by socketio for POST operations."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get('/{}'.format(API_NAME))
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                class_ = doc.parsed_classes[class_name]['class']
                class_methods = [x.method for x in class_.supportedOperation]
                if 'PUT' in class_methods:
                    # insert a object to be updated later
                    dummy_object = gen_dummy_object(class_.title, doc)
                    put_response = test_app_client.put(
                        endpoints[endpoint], data=json.dumps(dummy_object))
                    # extract id of the created object from the description of response
                    desc = put_response.json['description']
                    id_ = desc.split('ID ')[1].split(' successfully')[0]
                    # Flush old socketio updates
                    socketio_client.get_received('/sync')
                    # POST object
                    if 'POST' in class_methods:
                        new_dummy_object = gen_dummy_object(class_.title, doc)
                        post_response = test_app_client.post(
                            f'{endpoints[endpoint]}/{id_}',
                            data=json.dumps(new_dummy_object))
                        assert post_response.status_code == 200
                        # Get new socketio update
                        update = socketio_client.get_received('/sync')
                        assert len(update) != 0
                        assert update[0]['args'][0]['method'] == 'POST'
                        resource_name = update[0]['args'][0]['resource_url'].split('/')[-2]
                        assert resource_name == endpoints[endpoint].split('/')[-1]

    def test_socketio_DELETE_updates(self, socketio_client, test_app_client, constants, doc):
        """Test 'update' event emitted by socketio for DELETE operations."""
        API_NAME = constants['API_NAME']
        index = test_app_client.get('/{}'.format(API_NAME))
        assert index.status_code == 200
        endpoints = json.loads(index.data.decode('utf-8'))
        for endpoint in endpoints:
            if endpoint not in ['@context', '@id', '@type', 'collections']:
                class_name = '/'.join(endpoints[endpoint].split(f'/{API_NAME}/')[1:])
                class_ = doc.parsed_classes[class_name]['class']
                class_methods = [x.method for x in class_.supportedOperation]
                if 'PUT' in class_methods:
                    # insert a object first to be deleted later
                    dummy_object = gen_dummy_object(class_.title, doc)
                    put_response = test_app_client.put(
                        endpoints[endpoint], data=json.dumps(dummy_object))
                    # extract id of the created object from the description of response
                    desc = put_response.json['description']
                    id_ = desc.split('ID ')[1].split(' successfully')[0]
                    # Flush old socketio updates
                    socketio_client.get_received('/sync')
                    if 'DELETE' in class_methods:
                        delete_response = test_app_client.delete(f'{endpoints[endpoint]}/{id_}')
                        assert delete_response.status_code == 200
                        # Get new update event
                        update = socketio_client.get_received('/sync')
                        assert len(update) != 0
                        assert update[0]['args'][0]['method'] == 'DELETE'
                        resource_name = update[0]['args'][0]['resource_url'].split('/')[-2]
                        assert resource_name == endpoints[endpoint].split('/')[-1]



from typing import Dict, Any, List, Optional, Union, Tuple

from flask import Response, jsonify

import re

from hydrus.data import crud

from hydrus.utils import get_doc, get_api_name, get_hydrus_server_url, get_session
from hydrus.utils import get_collections_and_parsed_classes
from hydra_python_core.doc_writer import HydraIriTemplate, IriTemplateMapping, HydraLink
from hydra_python_core.doc_writer import HydraError, DocUrl
from hydrus.socketio_factory import socketio
from hydrus.conf import get_host_domain


def validObject(object_: Dict[str, Any]) -> bool:
    """
    Check if the Dict passed in POST is of valid format or not.
        (if there's an "@type" key in the dict)
    :param object_ - Object to be checked
    :return : <bool> True if Object has "@type" key
    """
    if "@type" in object_:
        return True
    return False


def validObjectList(objects_: List[Dict[str, Any]]) -> bool:
    """
    Check if the List of Dicts passed are of the valid format or not.
        (if there's an "@type" key in the dict)
    :param objects_: Object to be checked
    :return : <bool> True if all the Object in the List of Dicts have "@type" key
    """
    for object_ in objects_:
        if "@type" not in object_:
            return False
    return True


def get_iri_from_int_list(path_url: str, int_list: List[str]) -> List[str]:
    """Returns a list of full IRIs of multiple objects from int_list
    :param path_url: The path of the object
    :param int_list: Optional String containing ',' separated ID's
    :return: IRIs of multiple objects
    """
    iri_list = []
    for id_ in int_list:
        iri_list.append(f"{path_url}/{id_}")
    return iri_list


def type_match(object_: List[Dict[str, Any]], obj_type: str) -> bool:
    """
    Checks if the object type matches for every object in list.
    :param object_: List of objects
    :param obj_type: The required object type
    :return: True if all object of list have the right type
            False otherwise
    """
    for obj in object_:
        if obj["@type"] != obj_type:
            return False
    return True


def set_response_headers(resp: Response,
                         ct: str = "application/ld+json",
                         headers: List[Dict[str, Any]]=[],
                         status_code: int = 200) -> Response:
    """
    Set the response headers.
    :param resp: Response.
    :param ct: Content-type default "application/ld+json".
    :param headers: List of objects.
    :param status_code: status code default 200.
    :return: Response with headers.
    """
    resp.status_code = status_code
    for header in headers:
        resp.headers[list(header.keys())[0]] = header[list(header.keys())[0]]
    resp.headers['Content-type'] = ct
    link = "http://www.w3.org/ns/hydra/core#apiDocumentation"
    vocab_route = get_doc().doc_name
    link_header = f'<{get_hydrus_server_url()}{get_api_name()}/{vocab_route}>; rel="{link}"'
    resp.headers['Link'] = link_header
    return resp


def hydrafy(object_: Dict[str, Any], path: Optional[str]) -> Dict[str, Any]:
    """
    Add hydra context to objects.
    :param object_ : Object.
    :param path : Path of the collection or non-collection class .
    :return : object with hydra context
    """
    if path == object_["@type"]:
        object_[
            "@context"] = f"{get_host_domain()}/{get_api_name()}/contexts/{object_['@type']}.jsonld"
    else:
        object_[
            "@context"] = f"{get_host_domain()}/{get_api_name()}/contexts/{path}.jsonld"
    return object_


def checkEndpoint(method: str, path: str) -> Dict[str, Union[bool, int]]:
    """
    Check if endpoint and method is supported in the API.
    :param method: Method name
    :param path: Path of the collection or non-collection class
    :return : Dict with 'method' and 'status' key
    """
    status_val = 404
    vocab_route = get_doc().doc_name
    if path == vocab_route:
        return {'method': False, 'status': 405}
    expanded_base_url = DocUrl.doc_url
    for endpoint in get_doc().entrypoint.entrypoint.supportedProperty:
        expanded_endpoint_id = endpoint.id_.replace("EntryPoint/", "")
        endpoint_id = expanded_endpoint_id.split(expanded_base_url)[1]
        if path == endpoint_id:
            status_val = 405
            for operation in endpoint.supportedOperation:
                if operation.method == method:
                    status_val = 200
                    return {'method': True, 'status': status_val}
    return {'method': False, 'status': status_val}


def getType(class_path: str, method: str) -> Any:
    """
    Return the @type of object allowed for POST/PUT.
    :param class_path: path for the class
    :param method: Method name
    """
    expanded_base_url = DocUrl.doc_url
    for supportedOp in get_doc(
    ).parsed_classes[class_path]["class"].supportedOperation:
        if supportedOp.method == method:
            class_type = supportedOp.expects.split(expanded_base_url)[1]
            return class_type


def checkClassOp(path: str, method: str) -> bool:
    """
    Check if the Class supports the operation.
    :param path: Path of the collection or non-collection class.
    :param method: Method name.
    :return: True if the method is defined, false otherwise.
    """
    collections, parsed_classes = get_collections_and_parsed_classes()
    if path in collections:
        supported_operations = get_doc().collections[path]["collection"].supportedOperation
    else:
        supported_operations = get_doc().parsed_classes[path]["class"].supportedOperation
    for supportedOp in supported_operations:
        if supportedOp.method == method:
            return True
    return False


def check_required_props(path: str, obj: Dict[str, Any]) -> bool:
    """
    Check if the object contains all required properties.
    :param path: Path of the collection or non-collection class.
    :param obj: object under check
    :return: True if the object contains all required properties
             False otherwise.
    """
    collections, parsed_classes = get_collections_and_parsed_classes()
    if path in collections:
        # path is of a collection class
        supported_properties = get_doc().collections[path]["collection"].supportedProperty
    else:
        # path is of a non-collection class
        supported_properties = get_doc().parsed_classes[path]["class"].supportedProperty
    for prop in supported_properties:
        if prop.required:
            if prop.title not in obj:
                return False
    return True


def check_writeable_props(path: str, obj: Dict[str, Any]) -> bool:
    """
    Check that the object only contains writeable fields(properties).
    :param path: Path of the collection or non-collection class.
    :param obj: object under check
    :return: True if the object only contains writeable properties
             False otherwise.
    """
    collections, parsed_classes = get_collections_and_parsed_classes()
    if path in collections:
        # path is of a collection class
        supported_properties = get_doc().collections[path]["collection"].supportedProperty
    else:
        # path is of a non-collection class
        supported_properties = get_doc().parsed_classes[path]["class"].supportedProperty
    for prop in supported_properties:
        if prop.write is False:
            if prop.title in obj:
                return False
    return True


def get_nested_class_path(class_type: str) -> Tuple[str, bool]:
    """
    Get the path of class
    :param class_type: class name whose path is needed
    :return: Tuple, where the first element is the path string and
             the second element is a boolean, True if the class is a collection class
             False otherwise.
    """
    expanded_base_url = DocUrl.doc_url
    collections, parsed_classes = get_collections_and_parsed_classes()
    for path in collections:
        collection = collections[path]["collection"]
        class_name = collection.manages["object"].split(expanded_base_url)[1]
        collection_manages_class = parsed_classes[class_name]["class"]
        collection_manages_class_type = collection_manages_class.title
        collection_manages_class_path = collection_manages_class.path
        if collection_manages_class_type == class_type:
            return collection_manages_class_path, True
    for class_path in parsed_classes:
        if class_type == parsed_classes[class_path]["class"].title:
            return class_path, False


def finalize_response(path: str, obj: Dict[str, Any]) -> Dict[str, Any]:
    """
    finalize response objects by removing properties which are not readable and correcting path
    of nested objects.
    :param path: Path of the collection or non-collection class.
    :param obj: object being finalized
    :return: An object not containing any `readable=False` properties and having proper path
             of any nested object's url.
    """
    collections, parsed_classes = get_collections_and_parsed_classes()
    expanded_base_url = DocUrl.doc_url
    if path in collections:
        members = list()
        for member in obj["members"]:
            member_id = member[0]
            member_type = member[1]
            member_path = get_path_from_type(member_type)
            member = {
                "@type": "hydra:Link",
                "@id": f"{get_host_domain()}/{get_api_name()}/{member_path}/{member_id}",
            }
            members.append(member)
        obj['members'] = members
        return obj
    else:
        # path is of a non-collection class
        supported_properties = get_doc().parsed_classes[path]["class"].supportedProperty
        expanded_base_url = DocUrl.doc_url
        for prop in supported_properties:
            # Skip not required properties which are not inserted yet.
            if not prop.required and prop.title not in obj:
                continue
            # if prop.read is False:
            #     obj.pop(prop.title, None)
            elif isinstance(prop.prop, HydraLink):
                hydra_link = prop.prop
                range_class = hydra_link.range.split(expanded_base_url)[1]
                nested_path, is_collection = get_nested_class_path(range_class)
                if is_collection:
                    id = obj[prop.title]
                    obj[prop.title] = f"/{get_api_name()}/{nested_path}/{id}"
                else:
                    obj[prop.title] = f"/{get_api_name()}/{nested_path}"
            elif expanded_base_url in prop.prop:
                prop_class = prop.prop.split(expanded_base_url)[1]
                prop_class_path = parsed_classes[prop_class]['class'].path
                id = obj[prop.title]
                class_resp = crud.get(id, prop_class, get_api_name(), get_session(),
                                      path=prop_class_path)
                obj[prop.title] = finalize_response(prop_class_path, class_resp)
        return obj


def add_iri_template(path: str, API_NAME: str, collection_path: str) -> Dict[str, Any]:
    """
    Creates an IriTemplate.
    :param path: Path of the collection or the non-collection class.
    :param API_NAME: Name of API.
    :return: Hydra IriTemplate .
    """
    template_mappings = list()
    template = f"/{API_NAME}/{collection_path}{{?"
    template, template_mappings = generate_iri_mappings(path, template,
                                                        template_mapping=template_mappings,)

    template, template_mappings = add_pagination_iri_mappings(template=template,
                                                              template_mapping=template_mappings)
    return HydraIriTemplate(template=template, iri_mapping=template_mappings).generate()


def generate_iri_mappings(path: str, template: str, skip_nested: bool = False,
                          template_mapping: List[IriTemplateMapping] = [],
                          parent_prop_name: str = None) -> Tuple[str, List[IriTemplateMapping]]:
    """
    Generate iri mappings to add to IriTemplate
    :param path: Path of the collection or non-collection class.
    :param template: IriTemplate string.
    :param skip_nested: To only add properties of the class_type class or
                        its immediate children.
    :param template_mapping: List of template mappings.
    :param parent_prop_name: Property name according to parent object
                             (only applies for nested properties)
    :return: Template string, list of template mappings and boolean showing whether
             to keep adding delimiter or not.
    """
    expanded_base_url = DocUrl.doc_url
    for supportedProp in get_doc(
    ).parsed_classes[path]["class"].supportedProperty:
        prop_class = supportedProp.prop
        nested_class_prop = False
        if isinstance(supportedProp.prop, HydraLink):
            hydra_link = supportedProp.prop
            prop_class = hydra_link.range.split(expanded_base_url)[1]
            nested_class_prop = True
        elif expanded_base_url in supportedProp.prop:
            prop_class = supportedProp.prop.split(expanded_base_url)[1]
            nested_class_prop = True
        if nested_class_prop and skip_nested is False:
            template, template_mapping = generate_iri_mappings(prop_class, template,
                                                               skip_nested=True,
                                                               parent_prop_name=supportedProp.title,
                                                               template_mapping=template_mapping)
            continue
        if skip_nested is True:
            var = f"{parent_prop_name}[{supportedProp.title}]"
            mapping = IriTemplateMapping(variable=var, prop=prop_class)
        else:
            var = supportedProp.title
            mapping = IriTemplateMapping(variable=var, prop=prop_class)
        template_mapping.append(mapping)
        template = template + f"{var}, "
    return template, template_mapping


def add_pagination_iri_mappings(template: str,
                                template_mapping: List[IriTemplateMapping]
                                ) -> Tuple[str, List[IriTemplateMapping]]:
    """
    Add various pagination related to variable to the IRI template and also adds mappings for them.
    :param template: IriTemplate string.
    :param template_mapping: List of template mappings.
    :return: Final IriTemplate string and related list of mappings.
    """
    paginate_variables = ["pageIndex", "limit", "offset"]
    for i in range(len(paginate_variables)):
        # If final variable then do not add space and comma and add the final parentheses
        if i == len(paginate_variables) - 1:
            template += f"{paginate_variables[i]}}}"
        else:
            template += f"{paginate_variables[i]}, "
        mapping = IriTemplateMapping(variable=paginate_variables[i], prop=paginate_variables[i])
        template_mapping.append(mapping)
    return template, template_mapping


def send_sync_update(socketio, new_job_id: int, last_job_id: str,
                     method: str, resource_url: str):
    """
    Sends synchronization update to all connected clients.
    :param socketio: socketio connection.
    :param new_job_id: Job id of the new modification(update).
    :param last_job_id: Job id of the last(most recent) modification until this new one.
    :param method: Method type of the operation.
    :param resource_url: URL of resource which needs to be synchronized.
    """
    data = {
        "job_id": new_job_id,
        "last_job_id": last_job_id,
        "method": method,
        "resource_url": resource_url
    }
    socketio.emit('update', data, namespace="/sync")


def get_link_props(path: str, object_) -> Tuple[Dict[str, Any], bool]:
    """
    Get dict of all hydra_link properties of a class.
    :param path: Path of the collection or non-collection class.
    :param object_: Object being inserted/updated.
    :return: Tuple with one elements as Dict with property_title as key and
             instance_id(for collection class) or class_name(for non-collection class) as value,
             second element represents boolean representing validity of the link.
    """
    link_props = {}
    collections, parsed_classes = get_collections_and_parsed_classes()
    expanded_base_url = DocUrl.doc_url
    if path in collections:
        # path is of a collection class
        supported_properties = get_doc().collections[path]["collection"].supportedProperty
    else:
        # path is of a non-collection class
        supported_properties = get_doc().parsed_classes[path]["class"].supportedProperty
    for supportedProp in supported_properties:
        if isinstance(supportedProp.prop, HydraLink) and supportedProp.title in object_:
            prop_range = supportedProp.prop.range
            range_class_name = prop_range.split(expanded_base_url)[1]
            for collection_path in get_doc().collections:
                if collection_path in object_[supportedProp.title]:
                    class_title = get_doc().collections[collection_path]['collection'].class_.title
                    if range_class_name != class_title:
                        return {}, False
                    link_props[supportedProp.title] = object_[supportedProp.title].split('/')[-1]
                    break
            if supportedProp.title not in link_props:
                for class_path in get_doc().parsed_classes:
                    if class_path in object_[supportedProp.title]:
                        class_title = get_doc().parsed_classes[class_path]['class'].title
                        if range_class_name != class_title:
                            return {}, False
                        link_props[supportedProp.title] = class_title
                        break
    return link_props, True


def get_link_props_for_multiple_objects(path: str,
                                        object_list: List[Dict[str, Any]]
                                        ) -> Tuple[List[Dict[str, Any]], bool]:
    """
    Get link_props of multiple objects.
    :param path: Path of the collection or non-collection class.
    :param object_list: List of objects being inserted.
    :return: List of link properties processed with the help of get_link_props.
    """
    link_prop_list = list()
    for object_ in object_list:
        link_props, type_check = get_link_props(path, object_)
        if type_check is True:
            link_prop_list.append(link_props)
        else:
            return [], False
    return link_prop_list, True


def validate_object(object_: Dict[str, Any],
                    obj_type: str, class_path: str) -> bool:
    """
    Check if the object dict passed in POST can be inserted/updated
    in database.

    :param object_: Object to be checked
    :param obj_type: The required object type
    :param class_path: Path of the class
    :return: True if the object is completely valid
    """
    return (validObject(object_) and
            object_["@type"] == obj_type and
            check_required_props(class_path, object_))


def get_context(category: str) -> Response:
    """
    Generate the context for a given category.

    :param category: The category of class for which context is required
    :type category: str
    :return: Response with context
    :rtype: Response
    """
    collections, parsed_classes = get_collections_and_parsed_classes()
    # Check for collection
    if category in get_doc().collections:
        # type: Union[Dict[str,Any],Dict[int,str]]
        response = {"@context": collections[category]["context"].generate()}
        return set_response_headers(jsonify(response))
    # Check for non collection class
    elif category in parsed_classes:
        response = {"@context": get_doc().parsed_classes[category]["context"].generate()}
        return set_response_headers(jsonify(response))
    else:
        error = HydraError(code=404, title="NOT FOUND", desc="Context not found")
        return error_response(error)


def error_response(error: HydraError) -> Response:
    """
    Generate the response if there is an error while performing any operation

    :param error: HydraError object which will help in generating response
    :type error: HydraError
    :return: Error response with appropriate status code
    :rtype: Response
    """
    return set_response_headers(jsonify(error.generate()),
                                status_code=error.code)


def send_update(method: str, path: str):
    """Handler for sending synchronization update to all connected clients.

    :param method: Method type of the operation.
    :type method: str
    :param path: Path to the Item collection to which update is made.
    :type path: str
    """
    resource_url = f"{get_hydrus_server_url()}{get_api_name()}/{path}"
    session = get_session()
    last_job_id = crud.get_last_modification_job_id(session)
    new_job_id = crud.insert_modification_record(method, resource_url, session)
    send_sync_update(socketio, new_job_id, last_job_id, method, resource_url)


def get_path_from_type(type_: str) -> str:
    _, parsed_classes = get_collections_and_parsed_classes()
    expanded_base_url = DocUrl.doc_url
    for class_name in parsed_classes:
        class_ = parsed_classes[class_name]['class']
        if type_ == class_.id_.split(expanded_base_url)[1]:
            return class_.path


def parse_collection_members(object_: dict) -> dict:
    """Parse the members of a collection to make it easier
    to insert in database.

    :param object_: The body of the request having object members
    :type object_: dict
    :return: Object with parsed members
    :rtype: dict
    """
    members = list()
    for member in object_['members']:
        # example member
        # {
        #     "@id": "/serverapi/LogEntry/aab38f9d-516a-4bb2-ae16-068c0c5345bd",
        #     "@type": "LogEntry"
        # }
        member_id = member['@id'].split('/')[-1]
        member_type = member['@type']
        if crud.item_exists(member_type, member_id, get_session()):
            members.append({
                "id_": member_id,
                "@type": member_type,
            })
        else:
            error = HydraError(code=400, title="Data is not valid")
            return error_response(error)
    object_['members'] = members
    return object_


def get_fragments(resource: str) -> dict:
    """Gets a fragment of the main hydra vocabulary.

    :param resource: Fragment specified in the request parameters
    :type resource: str
    :return: Object referred to by the fragment
    :rtype: dict
    """
    resource_dict = dict()
    gen_doc = get_doc().generate()
    if 'EntryPoint/' in resource:
        match_string = r'\b(EntryPoint)\b'
        for class_ in gen_doc['supportedClass']:
            if re.search(match_string, class_['@id']):
                res = class_
                break
        for properties in res["supportedProperty"]:
            if resource in properties["property"]["@id"]:
                res = properties
                break
        resource_dict[resource[11:]] = res
    else:
        resource_dict['@context'] = gen_doc['@context']
        for class_ in gen_doc['supportedClass']:
            match_string = r'\b({0})\b'.format(resource)
            if re.search(match_string, class_['@id']):
                res = class_
                break
        resource_dict["supportedClass"] = [res]
    return resource_dict
